# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius Killinger and Philipp J. Schubert
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip

import logging
import time
from functools import reduce

import numpy as np
import theano
import theano.tensor as T

from ..config import config
from . import computations
from .variables import VariableWeight, ConstantParam, VariableParam
from .graphutils import floatX, TaggedShape, as_floatX
from .node_basic import Node, Concat

logger = logging.getLogger('elektronn2log')

__all__ = ['Perceptron', 'Conv', 'UpConv', 'Crop', 'LSTM',
           'FragmentsToDense', 'Pool', 'Dot', 'FaithlessMerge',
           'GRU', 'LRN', 'ImageAlign', 'UpConvMerge']

################################################################################

### TODO gradnet stuff anpassen von Conv layer in den anderen layers? Vlt nicht jetzt...

class NeuralLayer(Node):
    """
    Dummy class to add parameter initialisation methods for neural layers.
    """
    def _register_param(self, param, shape, name, init_kwargs=None,
                       apply_train=False, apply_reg=False):
        """
        Create parameter, set parameter as attribute and add to self.params if
        not shared from another Layer.

        Parameters
        ----------
        param: None or np.ndarray or T.Variable or list
            Possible forms of ``param``:
            * Passing ``None`` creates new parameter with default
              initialisation.
            * Passing a np.ndarray creates new parameter with the values
              of the array as initialisation.
            * A shared parameter is created by passing a T.Variable as
              ``param``.
            * A constant parameter is created by passing [np.ndarray, 'const']
              as ``param``.
              This parameter cannot be changed (no set_value) but makes the
              compiled function faster.
        shape: tuple
            Shape of the new the parameter (VariableWeight).
        name: str
            Parameter name.
        init_kwargs
            kwargs for utils.initialisation.
        apply_train: bool
            Train flag of the new parameter (VariableWeight).
        apply_reg: bool
            Regularisation flag of the new parameter (VariableWeight).
        """
        add_to_params = True
        if self.name=='':
            p_name = '<%s%s>'%(name, tuple(shape))
        else:
            p_name = '<%s_%s%s>'%(self.name, name, tuple(shape))
        # create new trainable by initialistaion
        if param is None:
            p = VariableWeight(shape=shape,
                               init_kwargs=init_kwargs,
                               name=p_name,
                               apply_train=apply_train,
                               apply_reg=apply_reg,)

        # create new trainable from values
        elif isinstance(param, np.ndarray):
            if param.shape!=tuple(shape):
                if not (param.ndim==0 and shape==(1,)):
                    raise ValueError("Shape mismatch. Required %s, given %s"\
                                 %(shape, param.shape))
            p = VariableWeight(value=param,
                               name=p_name,
                               apply_train=apply_train,
                               apply_reg=apply_reg,
                               dtype=floatX)

        # share a variable from elsewhere, not trainable
        elif isinstance(param, T.Variable): # (elektronn2.tensor.variables are T.Variable)
            try:
                sh = param.get_value().shape
                if sh!=tuple(shape):
                    raise ValueError("Shape mismatch. Required %s, given %s" \
                                     % (shape, param.shape))
            except AttributeError:
                logger.warning("Could not check correct shape of given weight %s, "
                               "make sure it has shape %s" %(param, shape))
            p = param
            add_to_params = False

        # create constant variable (or explicitly trainable
        elif isinstance(param, (list, tuple)):
            fail = False
            if not isinstance(param[0] , np.ndarray):
                fail = True
            if param[0].shape!=tuple(shape):
                raise ValueError("Shape mismatch. Required %s, given %s"\
                                 %(shape, param[0].shape))
            if param[1] == 'const':
                value = as_floatX(param[0])
                p = ConstantParam(value, p_name)
            elif param[1] == 'trainable':
                value = as_floatX(param[0])
                p = VariableWeight(value=value,
                                   name=p_name,
                                   apply_train=True,
                                   apply_reg=apply_reg)
            else:
                fail = True

            if fail:
                raise ValueError("If a parameter is passed as a list, the "
                                 "first entry must contain the parameter "
                                 "value (np.ndarray) and the second entry "
                                 "must be either 'const' or 'trainable' "
                                 "to indicate whether this param is "
                                 "trainable. Got [%s, %s]" \
                                 %(type(param[0]), param[1]))
        else:
            raise ValueError("Parameter %s must be either <np.ndarray>, "
                             "<theano.TensorVariable>, a tuple or None "
                             "(to create new param)" %(name,))

        setattr(self, name, p) #
        if add_to_params:
            self.params[name] = p
        else:
            logger.debug("Sharing theano variable %s. This parameter is not added to self.params" %(p,))


    def _setup_params(self, w_sh, w, b, gamma, mean, std, dropout_rate,
                      pool_shape=None, gradnet_rate=None):
        """
        Register each parameter, choose appropriate initialisation.
        """
        # Dot/Conv/Bias Parameters #############################################
        self.w = None

        # TODO: Pass w_init mode from layer to setup_params
        if config.use_ortho_init or isinstance(self, GRU) or isinstance(self, LSTM):
            w_init = dict(scale='glorot', mode='ortho', pool=pool_shape,
                          spatial_axes=self.spatial_axes)
        else:
            w_init = dict(scale='glorot', mode='normal', pool=pool_shape,
                          spatial_axes=self.spatial_axes)

        self._register_param(w, w_sh, 'w', init_kwargs=w_init,
                             apply_train=True, apply_reg=True)

        activation_func = self.activation_func
        n_f = self.n_f
        self.b = None
        if isinstance(self, GRU):
            b_sh=(3 * n_f, )
        elif isinstance(self, LSTM):
            b_sh = (4 * n_f, )
        else:
            b_sh=(n_f,)
        if activation_func=='relu' or activation_func.startswith("maxout"):
            norm = 1.0
            if len(w_sh) > 2:
                fov = 1
                for i in self.spatial_axes:
                    fov = fov * w_sh[i]
                norm = fov

            b_init=dict(scale=1.0/norm, mode='const')

        elif activation_func=='sigmoid':
            b_init=dict(scale=0.5, mode='const')
        elif activation_func=='prelu':
            norm = 1.0
            if len(w_sh) > 2:
                fov = 1
                for i in self.spatial_axes:
                    fov = fov * w_sh[i]
                norm = fov

            b_init=dict(scale=1.0/norm, mode='prelu')
            if isinstance(self, GRU):
                 b_sh=(3 * n_f, 2)
            elif isinstance(self, LSTM):
                b_sh = (4 * n_f, 2)
            else:
                b_sh=(n_f, 2)
        else: # all other activations
            b_init=dict(scale=1e-6, mode='fix-uni')
        self._register_param(b, b_sh, 'b', init_kwargs=b_init,
                            apply_train=True, apply_reg=False)

        # Batch Normalisation ##################################################
        batch_normalisation = self.batch_normalisation
        if batch_normalisation in ['train', 'fadeout']:
            # mean and std are created as TensorVariables in _calc_output
            self.gamma = None
            sh = (n_f,)
            g_init =dict(scale=1.0, mode='const')
            self._register_param(gamma, sh, 'gamma', init_kwargs=g_init,
                                apply_train=True, apply_reg=3.0) ###TODO maybe even stronger reg for this?
            if mean is not None or std is not None:
                raise ValueError("Cannot pass mean and std for training, they "
                                 "are computed in the theano graph.")

            # create mean and std for training to accumulate running avgs
            self.mean = None
            m_init =dict(scale=0.0, mode='const')
            self._register_param(None, sh, 'mean', init_kwargs=m_init)

            self.std = None
            s_init =dict(scale=1.0, mode='const')
            self._register_param(None, sh, 'std', init_kwargs=s_init)

        elif batch_normalisation=='predict':
            sh = (n_f,)
            self.gamma = None
            g_init =dict(scale=1.0, mode='const')
            self._register_param(gamma, sh, 'gamma', init_kwargs=g_init)

            self.mean = None
            m_init =dict(scale=0.0, mode='const')
            self._register_param(mean, sh, 'mean', init_kwargs=m_init)

            self.std = None
            s_init =dict(scale=1.0, mode='const')
            self._register_param(std, sh, 'std', init_kwargs=s_init)
        else:
            if batch_normalisation is not False:
                raise ValueError("Unknown value %s for batchnormalisation" %batch_normalisation)

        # Dropout ##############################################################
        self.dropout_rate = None
        if dropout_rate:
            value = as_floatX(dropout_rate)
            self._register_param(value, (1,), 'dropout_rate')


        # GradNet ##############################################################
        self.gradnet_rate = None
        if gradnet_rate:
            value = as_floatX(gradnet_rate)
            self._register_param(value, (1,), 'gradnet_rate')

###############################################################################

class Perceptron(NeuralLayer):
    """
    Perceptron Layer.

    Parameters
    ----------
    parent: Node or list of Node
        The input node(s).
    n_f: int
        Number of filters (nodes) in layer.
    activation_func: str
        Activation function name.
    flatten: bool
    batch_normalisation: str or None
        Batch normalisation mode.
        Can be False (inactive), "train" or "fadeout".
    dropout_rate: float
        Dropout rate (probability that a node drops out in a training step).
    name: str
        Perceptron name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w: np.ndarray or T.TensorVariable
        Weight matrix.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    b: np.ndarray or T.TensorVariable
        Bias vector.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    gamma
        (For batch normalisation) Initializes gamma parameter.
    mean
        (For batch normalisation) Initializes mean parameter.
    std
        (For batch normalisation) Initializes std parameter.
    gradnet_mode
    """  # TODO: Write docs on batch normalisation modes.
    # TODO: gradnet_mode seems to be unused. Should it be removed?

    def __init__(self, parent, n_f, activation_func='relu',
                 flatten=False, batch_normalisation=False, dropout_rate=0,
                 name="dot", print_repr=True, w=None, b=None, gamma=None,
                 mean=None, std=None, gradnet_mode=None):
        super(Perceptron, self).__init__(parent, name, print_repr)

        self.n_f = n_f
        self.activation_func = activation_func
        self.batch_normalisation = batch_normalisation
        self.gradnet_mode = gradnet_mode
        self.axis = parent.shape.tag2index('f') #retrieve feature shape's index
        self.flatten = flatten
        self.spatial_axes = parent.shape.spatial_axes
        if flatten:
            n_in = parent.shape.stripbatch_prod
        else:
            n_in = parent.shape['f'] #retrieve feature shape

        w_sh = (n_in, n_f)
        self._setup_params(w_sh, w, b, gamma, mean, std, dropout_rate)


    def _make_output(self):
        """
        Computation of Theano output.
        """
        if self.flatten:
            if self.axis is not 1:
                raise NotImplementedError("Cannot flatten tensor for "
                                          "Perceptron layer when batchsize is "
                                          "not on first axis")
            input_tensor = self.parent.output.flatten(2)
            pattern  = ['x', 0]
        else:
            input_tensor = self.parent.output
            pattern  = ['x' for i in input_tensor.shape]
            pattern[self.axis] = 0

        activation_func = self.activation_func
        if activation_func.startswith("maxout"):
            r=int(activation_func.split(" ")[1])
            assert r>=2
            self.n_f /= r

        if activation_func=='prelu':
            b  = self.b[:,0].dimshuffle(pattern)
            b1 = self.b[:,1].dimshuffle(pattern)
        else:
            b   = self.b.dimshuffle(pattern)
            b1  = None

        lin_output = computations.dot(input_tensor, self.w, self.axis)

        if self.batch_normalisation in ['train', 'fadeout']:
            mean = computations.apply_except_axis(
                lin_output,self.axis, T.mean).dimshuffle(pattern)
            std = computations.apply_except_axis(
                lin_output,self.axis, T.std).dimshuffle(pattern) + 1e-6
            gamma = self.gamma.dimshuffle(pattern)

            if self.batch_normalisation=='fadeout':
                logger.warning("Batch normalisation mode 'fadeout' does not "
                               "work for less than 50%...")
                mean = self.gradnet_rate * mean
                std  = self.gradnet_rate * std + (1-self.gradnet_rate) * 1.0
                gamma = self.gradnet_rate * gamma

            self.mean.updates = (self.mean,
                                 0.9995 * self.mean + 0.0005 * T.extra_ops.squeeze(mean))
            self.std.updates  = (self.std,
                                 0.9995 * self.std + 0.0005 * T.extra_ops.squeeze(std))

        elif self.batch_normalisation=='predict':
            mean = self.mean.dimshuffle(pattern)
            std = self.std.dimshuffle(pattern)
            gamma = self.gamma.dimshuffle(pattern)
        else:
            mean = 0
            std = 1
            gamma = 1




        lin_output =  (gamma / std) * lin_output + b - (gamma * mean / std)
        lin_output = computations.apply_activation(lin_output, activation_func, b1)

        if self.dropout_rate:
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p   = 1 - self.dropout_rate
            dropout_gate = rng.binomial(size=(self.n_f,), n=1, p=p,
                                        dtype=theano.config.floatX)
            dropout_gate *= 1.0 / p
            lin_output =  lin_output * dropout_gate.dimshuffle(pattern)

        self.output = lin_output


    def _calc_shape(self):
        """
        Calculate shape from parent shape and n_f and set it as self.shape.
        """
        sh = self.parent.shape
        if self.flatten:
            self.shape = TaggedShape((sh['b'], self.n_f), 'b,f')
        else:
            self.shape = sh.updateshape('f', self.n_f)


    def _calc_comp_cost(self):
        """
        Calculate abstract computational cost from parent shape and n_f and
        set it as self.computational_cost.
        """
        n = self.parent.shape.stripnone_prod
        self.computational_cost = n * self.n_f


    def make_dual(self, parent, share_w=False, **kwargs):
        """
        Create the inverse of this ``Perceptron``.

        Most options are the same as for the layer itself.
        If ``kwargs`` are not specified, the values of the primal
        layers are re-used and new parameters are created.

        Parameters
        ----------
        parent: Node
            The input node.
        share_w: bool
            If the weights (``w``) should be shared from the primal layer.
        kwargs: dict
            kwargs that are passed through to the constructor of the inverted
            Perceptron (see signature of ``Perceptron``).
            ``n_f`` is copied from the existing node on
            which ``make_dual`` is called.
            Every other parameter can be changed from the original
            ``Perceptron``'s defaults by specifying it in ``kwargs``.

        Returns
        -------
        Perceptron
            The inverted perceptron layer.
        """
        if self.flatten:
            raise NotImplementedError("Cannot make dual Layer for flattened "
                                      "Perceptron layer.")

        dropout_rate = 0.0 if not self.dropout_rate else self.dropout_rate.get_value()
        defaults = dict(activation_func=self.activation_func,
                        batch_normalisation=self.batch_normalisation,
                        dropout_rate=dropout_rate,
                        name=self.name+'.T',
                        print_repr=self._print_repr,
                        w=None, b=None,gamma=None, mean=None, std=None)

        defaults.update(kwargs)
        kwargs = defaults

        if share_w:
            if kwargs['w'] is not None:
                logger.debug("Ignoring passed w because w is shared from primal Layer.")
            kwargs['w'] = self.w.T

        n_f = self.parent.shape['f'] # This is the output of the dual Layer

        if self.n_f is not parent.shape['f']: # input of dual Layer  #q: Shouldn't this be "!=", instead of "is not"?
            raise ValueError("Cannot make dual layer of:\n"
                             "%s\n"
                             "with input: %s!\n"
                             "The output shape of the input for the dual layer "
                             "must match the the input shape of the primal layer."\
                             %(self, parent))

        return Perceptron(parent, n_f, **kwargs)


    def __repr__(self):
        s = super(NeuralLayer, self).__repr__()
        s += "\n"
        s += "  n_f=%i, " %(self.n_f,)
        s += "act='%s', " %(self.activation_func,)
        if self.flatten:
            s += "input was flattened, "
        if self.dropout_rate:
            s += "dropout rate = %.1f, "%(self.dropout_rate.get_value())
        if self.batch_normalisation:
            s += "BN in '%s' mode "%(self.batch_normalisation,)

        return s

Dot = Perceptron

###############################################################################

class Conv(Perceptron):
    """
    Convolutional layer with subsequent pooling.

    Examples
    --------
    Examples for constructing convolutional neural networks can be found
    in examples/neuro3d.py and examples/mnist.py.

    Parameters
    ----------
    parent: Node
        The input node.
    n_f: int
        Number of features.
    filter_shape: tuple
        Shape of the convolution filter kernels.
    pool_shape: tuple
        Size/shape of pooling after the convolution.
    conv_mode: str
        Possible values:
        * "valid": only apply filter to complete patches of the image.
          Generates output of shape: image_shape -filter_shape + 1.
        * "full" zero-pads image to multiple of filter shape to generate
          output of shape: image_shape + filter_shape - 1.
    activation_func: str
        Activation function name.
    mfp: bool
        Whether to apply Max-Fragment-Pooling in this Layer.
    batch_normalisation: str or None
        Batch normalisation mode.
        Can be False (inactive), "train" or "fadeout".
    dropout_rate: float
        Dropout rate (probability that a node drops out in a training step).
    name: str
        Layer name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w: np.ndarray or T.TensorVariable
        Weight matrix.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    b: np.ndarray or T.TensorVariable
        Bias vector.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    gamma
        (For batch normalisation) Initializes gamma parameter.
    mean
        (For batch normalisation) Initializes mean parameter.
    std
        (For batch normalisation) Initializes std parameter.
    gradnet_mode
    """

    def __init__(self, parent, n_f, filter_shape, pool_shape,
                 conv_mode='valid', activation_func='relu',
                 mfp=False, batch_normalisation=False, dropout_rate=0,
                 name="conv", print_repr=True, w=None, b=None, gamma=None,
                 mean=None, std=None, gradnet_mode=None):
        super(Perceptron, self).__init__(parent, name, print_repr)

        self.n_f  = n_f
        self.filter_shape = filter_shape
        self.pool_shape = pool_shape
        self.conv_mode = conv_mode
        self.activation_func = activation_func
        self.batch_normalisation = batch_normalisation
        self.gradnet_mode = gradnet_mode
        self.mfp = mfp

        self.strides = parent.shape.strides
        self.mfp_offsets = parent.shape.mfp_offsets
        self.axis = parent.shape.tag2index('f') #retrieve feature shape's index
        self.axis_order = None

        self.spatial_axes = self.parent.shape.spatial_axes
        conv_dim = len(self.spatial_axes)
        x_dim    = len(self.parent.shape)
        if len(self.spatial_axes)!=len(filter_shape) or \
                        len(filter_shape)!=len(pool_shape):
            raise ValueError("The filter_shape dimensionality (%i), the number "
                             "of spatial dimensions in the input (%i) and "
                             "the dimensionality of pool_shape (%i) differ! "
                             "Use filter size 1 on axes which should not be "
                             "convolved."\
                             %(len(filter_shape), conv_dim, len(pool_shape)))

        n_in = parent.shape['f'] #retrieve feature shape
        fail = False
        if conv_dim==1:
            if x_dim!=3 or self.spatial_axes!=[2]:
                fail = True
            w_sh = [n_f, n_in] + list(filter_shape)

        elif conv_dim==2:
            if x_dim!=4 or self.spatial_axes!=[2,3]:
                fail = True
            w_sh = [n_f, n_in] + list(filter_shape)

        elif conv_dim==3:
            if x_dim!=5:
                fail = True
            if self.spatial_axes==[2,3,4]:
                self.axis_order = 'dnn'
                w_sh = [n_f, n_in] + list(filter_shape)
            elif self.spatial_axes==[1,3,4]:
                self.axis_order = 'theano'
                w_sh = [n_f, filter_shape[0], n_in] + list(filter_shape[1:])
            else:
                fail = True

        if fail:
            raise NotImplementedError("Cannot convolve non-standard shapes / axis orders. "
                                      "Implement reshaping before conv "
                                      "and re-reshaping after!")

        self.conv_dim = conv_dim
        self.w_sh = w_sh

        gradnet_rate = 1.0 if gradnet_mode else None

        self._setup_params(w_sh, w, b, gamma, mean, std, dropout_rate,
                           pool_shape, gradnet_rate)


    def _make_output(self):
        """
        Computation of Theano output.
        """
        input_tensor = self.parent.output
        input_shape  = list(self.parent.shape)
        pattern  = ['x' for i in input_tensor.shape]
        pattern[self.axis] = 0
        activation_func = self.activation_func
        if activation_func.startswith("maxout"):
            r=int(activation_func.split(" ")[1])
            assert r>=2
            self.filter_shape /= r

        if activation_func=='prelu':
            b  = self.b[:,0].dimshuffle(pattern)
            b1 = self.b[:,1].dimshuffle(pattern)
        else:
            b   = self.b.dimshuffle(pattern)
            b1  = None

        lin_output = computations.conv(input_tensor, self.w, self.axis_order,
                                       border_mode=self.conv_mode,
                                       x_shape=input_shape, w_shape=self.w_sh)

        if self.mfp:
            if self.input_nodes[0].shape['b']!=1:
                raise ValueError("For MFP the batchsize of the raw image input must be 1.")

            lin_output, offsets_new, strides_new = computations.fragmentpool(lin_output,
                                                                        self.pool_shape,
                                                                        self.mfp_offsets,
                                                                        self.strides,
                                                                        self.spatial_axes)
            self.mfp_offsets = offsets_new
            self.strides = strides_new
        else:
            lin_output = computations.pooling(lin_output, self.pool_shape, self.spatial_axes)
            self.strides = np.multiply(self.pool_shape, self.strides)

        if self.batch_normalisation in ['train', 'fadeout']:
            mean = computations.apply_except_axis(
                lin_output,self.axis, T.mean).dimshuffle(pattern)
            std = computations.apply_except_axis(
                lin_output,self.axis, T.std).dimshuffle(pattern) + 1e-6
            gamma = self.gamma.dimshuffle(pattern)

            if self.batch_normalisation=='fadeout':
                logger.warning("Batch Normalisation mode 'fadeout' does not "
                               "work for less than 50%...")
                mean = self.gradnet_rate * mean
                std  = self.gradnet_rate * std + (1-self.gradnet_rate) * 1.0
                gamma = self.gradnet_rate * gamma

            self.mean.updates = (self.mean,
                                 0.9995 * self.mean + 0.0005 * T.extra_ops.squeeze(mean))
            self.std.updates  = (self.std,
                                 0.9995 * self.std + 0.0005 * T.extra_ops.squeeze(std))

        elif self.batch_normalisation=='predict':
            mean = self.mean.dimshuffle(pattern)
            std = self.std.dimshuffle(pattern)
            gamma = self.gamma.dimshuffle(pattern)
        else:
            mean = 0
            std = 1
            gamma = 1



        lin_output =  (gamma / std) * lin_output + b - (gamma * mean / std)
        lin_output = computations.apply_activation(lin_output, activation_func, b1)

        if self.dropout_rate:
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p   = 1 - self.dropout_rate
            dropout_gate = rng.binomial(size=lin_output.shape, n=1, p=p,
                                        dtype=theano.config.floatX)
            dropout_gate *= 1.0 / p
            lin_output *= dropout_gate  #.dimshuffle(('x', 0))

        self.output = lin_output


    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent.shape
        for j,(i,f,p) in enumerate(zip(self.spatial_axes, self.filter_shape, self.pool_shape)):
            if self.conv_mode=='valid':
                k = 1 - f
            elif self.conv_mode=='full':
                k = f - 1
            elif self.conv_mode=='same':
                k = 0
            s = (sh[i] + k)//p
            if self.mfp:
                if (sh[i] + k - p + 1)%p!=0:
                    raise ValueError("Cannot pool spatial axis '%s' of length %i "
                                     "by factor %i after convolving with "
                                     "kernel of size %i and using MFP."\
                                     %(sh.tags[i], sh[i], p, f))
            else:
                if (sh[i] + k)%p!=0:
                    raise ValueError("Cannot pool spatial axis '%s' of length %i "
                                     "by factor %i after convolving with "
                                     "kernel of size %i."\
                                     %(sh.tags[i], sh[i], p, f))
            sh = sh.updateshape(i, s)
            if sh.fov[j]>0:
                fov = sh.fov[j] + (f+p-2) * sh.strides[j]
            else:
                fov = -1
            sh = sh.updatefov(j, fov)

        if self.mfp:
            sh = sh.updatemfp_offsets(self.mfp_offsets)
            sh = sh.updateshape('b', np.prod(self.pool_shape), mode='mult')

        sh = sh.updatestrides(self.strides)
        sh = sh.updateshape('f', self.n_f)
        self.shape = sh


    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.
        """
        sh = self.parent.shape
        n_position = 1
        for i,f,p in zip(self.spatial_axes, self.filter_shape, self.pool_shape):
            s = 1 - f if self.conv_mode=='valid' else f -1
            n_position *= sh[i] + s

        b = 1 if sh['b'] is None else sh['b']
        self.computational_cost = np.product(self.w_sh) * n_position * b


    def make_dual(self, parent, share_w=False, mfp=False, **kwargs):
        """
        Create the inverse (``UpConv``) of this ``Conv`` node.

        Most options are the same as for the layer itself.
        If ``kwargs`` are not specified, the values of the primal
        layers are re-used and new parameters are created.

        Parameters
        ----------
        parent: Node
            The input node.
        share_w: bool
            If the weights (``w``) should be shared from the primal layer.
        mfp: bool
            If max-fragment-pooling is used.
        kwargs: dict
            kwargs that are passed through to the new ``UpConv`` node (see
            signature of ``UpConv``).
            ``n_f`` and ``pool_shape`` are copied from the existing node on
            which ``make_dual`` is called.
            Every other parameter can be changed from the original
            ``Conv``'s defaults by specifying it in ``kwargs``.

        Returns
        -------
        UpConv
            The inverted conv layer (as an ``UpConv`` node).
        """
        if mfp:
            parent = FragmentsToDense(parent, print_repr=False)

        dropout_rate = 0.0 if not self.dropout_rate else self.dropout_rate.get_value()
        defaults = dict(conv_mode='valid', activation_func=self.activation_func,
                        batch_normalisation=self.batch_normalisation,
                        dropout_rate=dropout_rate,
                        name=self.name+'.T',
                        print_repr=self._print_repr,
                        w=None, b=None,gamma=None, mean=None, std=None)

        defaults.update(kwargs)
        kwargs = defaults

        if share_w:
            if kwargs['w'] is not None:
                logger.debug("Ignoring passed w because w is shared from primal Layer.")
            w = self.w
            # Exchange n_in and n_f
            swap = (0,2) if (self.conv_dim==3 and self.axis_order=='theano') else (0,1)
            w = T.swapaxes(w, *swap)
            kwargs['w'] = w

        n_f = self.parent.shape['f'] # This is the output of the dual Layer

        if self.w_sh[0] is not parent.shape['f']: # input of dual Layer
            raise ValueError("Cannot make dual layer of:\n"
                             "%s\n"
                             "with input: %s!\n"
                             "The output shape of the input for the dual layer "
                             "must match the the input shape of the primal layer."\
                             %(self, parent))

        return UpConv(parent, n_f, self.pool_shape, **kwargs)

    def __repr__(self):
        s = super(NeuralLayer, self).__repr__()
        s += "\n"
        s += "  n_f=%i, " %(self.n_f,)
        s += "%id conv, kernel=%s, pool=%s, "\
             %(self.conv_dim, self.filter_shape, self.pool_shape)
        s += "act='%s', " %(self.activation_func,)
        if self.dropout_rate:
            s += "Dropout rate=%.1f, "%(self.dropout_rate.get_value())
        if self.batch_normalisation:
            s += "BN in '%s' mode "%(self.batch_normalisation,)
        if self.mfp:
            s += "MFP active, "

        return s

###############################################################################

class FragmentsToDense(Node):
    def __init__(self, parent, name="to_dense", print_repr=True):
        super(FragmentsToDense, self).__init__(parent, name, print_repr)

    def _make_output(self):
        """
        Computation of Theano output.
        """
        fragments  = self.parent.output
        sh = self.parent.shape
        if sh['b']!=len(sh.mfp_offsets) or sh['b']!=np.prod(sh.strides):
            raise ValueError("Need %i fragments on the batch axis. "
                             "Is MFP active at all?" %np.prod(sh.strides))

        self.output = computations.fragments2dense(fragments, sh.mfp_offsets,
                                              sh.strides, sh.spatial_axes)

    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent.shape
        for ax, st in zip(sh.spatial_axes, sh.strides):
            sh = sh.updateshape(ax, st, mode='mult')

        sh = sh.updateshape('b', 1)
        new_strides = np.ones(len(sh.spatial_axes), np.int)
        new_offsets = np.zeros((1,len(sh.spatial_axes)), np.int)
        self.shape = TaggedShape(sh.shape, sh.tags, new_strides,
                                 new_offsets, sh.fov)


    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.

        For this Node type this is hard-coded to 0.
        """
        self.computational_cost = 0

###############################################################################

###############################################################################

class UpConv(Conv):
    """
    Upconvolution layer.

    E.g. pooling + upconv with p=3::

          x x x x x x x x x    before pooling (not in this layer)
           \|/   \|/   \|/     pooling (not in this layer)
            x     x     x      input to this layer
        0 0 x 0 0 x 0 0 x 0 0  unpooling + padding (done in this layer)
           /|\   /|\   /|\     conv on unpooled (done in this layer)
          y y y y y y y y y    result of this layer

    Parameters
    ----------
    parent: Node
        The input node.
    n_f: int
        Number of filters (nodes) in layer.
    pool_shape: tuple
        Size/shape of pooling.
    activation_func: str
        Activation function name.
    identity_init: bool
        Initialise weights to result in pixel repetition upsampling
    batch_normalisation: str or None
        Batch normalisation mode.
        Can be False (inactive), "train" or "fadeout".
    dropout_rate: float
        Dropout rate (probability that a node drops out in a training step).
    name: str
        Layer name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w: np.ndarray or T.TensorVariable
        Weight matrix.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    b: np.ndarray or T.TensorVariable
        Bias vector.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    gamma
        (For batch normalisation) Initializes gamma parameter.
    mean
        (For batch normalisation) Initializes mean parameter.
    std
        (For batch normalisation) Initializes std parameter.
    gradnet_mode
    """

    def __init__(self, parent, n_f, pool_shape, activation_func='relu',
                 identity_init=True, batch_normalisation=False, dropout_rate=0,
                 name="upconv", print_repr=True, w=None, b=None, gamma=None,
                 mean=None, std=None, gradnet_mode=None):
        filter_shape = pool_shape
        super(UpConv, self).__init__(parent, n_f, filter_shape, pool_shape,
                                     'valid', activation_func, mfp=False,
                                     batch_normalisation=batch_normalisation,
                                     dropout_rate=dropout_rate, name=name,
                                     print_repr=print_repr, w=w, b=b,
                                     gamma=gamma, mean=mean, std=std,
                                     gradnet_mode=gradnet_mode)

        if identity_init:
            try:
                w_val = self.w.get_value() * 0.1
                s = np.minimum(w_val.shape[0], w_val.shape[1])
                s = np.arange(s)
                w_val[s,s] = 1.0
                self.w.set_value(w_val)
                self.b.set_value(self.b.get_value()*0.0)
            except:
                logger.warn("identity_init failed")


    def _make_output(self):
        """
        Computation of Theano output.
        """
        input_tensor = self.parent.output
        input_shape  = list(self.parent.shape)
        pattern  = ['x' for i in input_tensor.shape]
        pattern[self.axis] = 0
        activation_func = self.activation_func
        if activation_func.startswith("maxout"):
            r=int(activation_func.split(" ")[1])
            assert r>=2
            self.filter_shape /= r

        if activation_func=='prelu':
            b  = self.b[:,0].dimshuffle(pattern)
            b1 = self.b[:,1].dimshuffle(pattern)
        else:
            b   = self.b.dimshuffle(pattern)
            b1  = None

        spax = self.spatial_axes
        pool = np.array(self.pool_shape)
        input_shape_up = np.array(input_shape)
        if len(spax)==3 and not computations.dnn_avail:
          unpooled = computations.unpooling(input_tensor, self.pool_shape, self.spatial_axes)
          self._debug_outputs.append(unpooled)
          input_shape_up[spax] = input_shape_up[spax] * pool + pool - 1
          input_shape_up = list(input_shape_up)
          lin_output = computations.conv(unpooled, self.w, self.axis_order,
                                           border_mode=self.conv_mode,
                                           x_shape=input_shape_up, w_shape=self.w_sh)
        else:
            input_shape_up[spax] = input_shape_up[spax] * pool
            input_shape_up = list(input_shape_up)
            w = T.swapaxes(self.w, 0, 1)
            w_sh = list(self.w_sh)
            w_sh[0], w_sh[1] = w_sh[1], w_sh[0]
            lin_output = computations.upconv(input_tensor, w, self.pool_shape,
                                             x_shape=input_shape_up,
                                             w_shape=w_sh,
                                             axis_order='dnn')

        if self.batch_normalisation in ['train', 'fadeout']:
            mean = computations.apply_except_axis(
                lin_output,self.axis, T.mean).dimshuffle(pattern)
            std = computations.apply_except_axis(
                lin_output,self.axis, T.std).dimshuffle(pattern) + 1e-6
            gamma = self.gamma.dimshuffle(pattern)

            if self.batch_normalisation=='fadeout':
                logger.warning("Batch normalisation mode 'fadeout' does not "
                               "work for less than 50%...")
                mean = self.gradnet_rate * mean
                std  = self.gradnet_rate * std + (1-self.gradnet_rate) * 1.0
                gamma = self.gradnet_rate * gamma

            self.mean.updates = (self.mean,
                                 0.9995 * self.mean + 0.0005 * T.extra_ops.squeeze(mean))
            self.std.updates  = (self.std,
                                 0.9995 * self.std + 0.0005 * T.extra_ops.squeeze(std))

        elif self.batch_normalisation=='predict':
            mean = self.mean.dimshuffle(pattern)
            std = self.std.dimshuffle(pattern)
            gamma = self.gamma.dimshuffle(pattern)
        else:
            mean = 0
            std = 1
            gamma = 1


        lin_output =  (gamma / std) * lin_output + b - (gamma * mean / std)
        lin_output = computations.apply_activation(lin_output, activation_func, b1)

        if self.dropout_rate:
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p   = 1 - self.dropout_rate
            dropout_gate = rng.binomial(size=lin_output.shape, n=1, p=p,
                                        dtype=theano.config.floatX)
            dropout_gate *= 1.0 / p
            lin_output *= dropout_gate  #.dimshuffle(('x', 0))

        self.output = lin_output

    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        self.strides = np.divide(self.strides,self.pool_shape)
        sh = self.parent.shape
        for j,(i,f,p) in enumerate(zip(self.spatial_axes, self.filter_shape, self.pool_shape)):
            s = 1 - f if self.conv_mode=='valid' else f -1
            s = (sh[i] * p) + p - 1 + s # unpool with margin then apply conv
            sh = sh.updateshape(i, s)
            # Unpooling creates asymmetric FOV (left/right is different for
            # some neurons), therefore we flag the FOV as exceptional with '-1'
            sh = sh.updatefov(j, -1)

        sh = sh.updateshape('f', self.n_f)
        sh = sh.updatestrides(self.strides)
        self.shape = sh


    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.
        """
        sh = self.parent.shape
        n_position = 1
        for i,f,p in zip(self.spatial_axes, self.filter_shape, self.pool_shape):
            s = 1 - f if self.conv_mode=='valid' else f -1
            n_position *= (sh[i] * p) + s

        b = 1 if sh['b'] is None else sh['b']
        self.computational_cost = np.product(self.w_sh) * n_position * b



    def __repr__(self):
        s = super(NeuralLayer, self).__repr__()
        s += "\n"
        s += "  n_f=%i, " %(self.n_f,)
        s += "%id upconv, kernel=%s, pool=%s, "\
             %(self.conv_dim, self.filter_shape, self.pool_shape)
        s += "act='%s', " %(self.activation_func,)
        if self.dropout_rate:
            s += "Dropout rate=%.1f, "%(self.dropout_rate.get_value())
        if self.batch_normalisation:
            s += "BN in '%s' mode "%(self.batch_normalisation,)
        return s

    def make_dual(self, *args, **kwargs):
        raise NotImplementedError("Use Conv instead?")


class Crop(Node):
    """
    This node type crops the output of its parent.

    Parameters
    ----------
    parent: Node
        The input node whose output should be cropped.
    crop: tuple or list of ints
        Crop each spatial axis from either side by this number.
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """  # TODO: Write an example
    def __init__(self, parent, crop, name="crop", print_repr=False):

        super(Crop, self).__init__(parent, name, print_repr)
        self.crop=crop

    def _make_output(self):
        """
        Computation of Theano output.
        """
        # It is assumed that all other dimensions are matching
        cropper = []
        k = 0
        for i,s in enumerate(self.parent.shape):
            if i in self.parent.shape.spatial_axes:
                off = self.crop[k]
                cropper.append(slice(off, s-off))
                k += 1
            else:
                cropper.append(slice(None))

        cropper = tuple(cropper)
        self.output = self.parent.output[cropper]

    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent.shape.copy()
        k = 0
        for i,s in enumerate(self.parent.shape):
            if i in self.parent.shape.spatial_axes:
                off = self.crop[k]
                sh = sh.updateshape(i,s-2*off)
                k += 1

        self.shape = sh

    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.

        For this Node type this is hard-coded to 0.
        """
        self.computational_cost = 0


def ImageAlign(hi_res, lo_res, hig_res_n_f,
                    activation_func='relu', identity_init=True,
                    batch_normalisation=False, dropout_rate=0,
                    name="upconv", print_repr=True, w=None, b=None, gamma=None,
                    mean=None, std=None, gradnet_mode=None):
    """
    Try to automatically align and concatenate a high-res and a low-res
    convolution output of two branches of a CNN by applying UpConv and Crop to
    make their shapes and strides compatible.
    UpConv is used if the low-res Node's strides are at least twice as large
    as the strides of the high-res Node in any dimension.

    This function can be used to simplify creation of e.g. architectures similar to
    U-Net (see https://arxiv.org/abs/1505.04597).

    If a ValueError that the shapes cannot be aligned is thrown,
    you can try changing the filter shapes and pooling factors of the
    (grand-)parent Nodes or add/remove Convolutions and Crops in the preceding
    branches until the error disappears (of course you should try to keep
    those changes as minimal as possible).

    (This function is an alias for UpConvMerge.)

    Parameters
    ----------
    hi_res: Node
        Parent Node with high resolution output.
    lo_res: Node
        Parent Node with low resolution output.
    hig_res_n_f: int
        Number of filters for the aligning UpConv.
    activation_func: str
        (passed to new UpConv if required).
    identity_init: bool
        (passed to new UpConv if required).
    batch_normalisation: bool
        (passed to new UpConv if required).
    dropout_rate: float
        (passed to new UpConv if required).
    name: str
        Name of the intermediate UpConv node if required.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w
        (passed to new UpConv if required).
    b
        (passed to new UpConv if required).
    gamma
        (passed to new UpConv if required).
    mean
        (passed to new UpConv if required).
    std
        (passed to new UpConv if required).
    gradnet_mode
        (passed to new UpConv if required).

    Returns
    -------
    Concat
        Concat Node that merges the aligned high-res and low-res outputs.
    """
    ###TODO exchange UpConv and Crop to save computation in some cases

    sh_hi = hi_res.shape
    sh_lo = lo_res.shape
    assert len(sh_hi)==len(sh_lo)
    assert sh_hi.spatial_axes == sh_lo.spatial_axes

    unpool = sh_lo.strides // sh_hi.strides
    if np.any(unpool>1):
        lo_res = UpConv(lo_res, hig_res_n_f, unpool,
               activation_func=activation_func, identity_init=identity_init,
               batch_normalisation=batch_normalisation, dropout_rate=dropout_rate,
               name=name, print_repr=print_repr, w=w, b=b, gamma=gamma,
               mean=mean, std=std, gradnet_mode=gradnet_mode)

    # No both have same stride
    # Shapes may have changed
    sh_hi = hi_res.shape.spatial_shape
    sh_lo = lo_res.shape.spatial_shape

    crop_lo = []
    crop_hi = []
    for i in range(len(sh_hi)):
        diff = sh_hi[i] - sh_lo[i]  # different in orignal space
        if diff % 2!=0:
            raise ValueError("hi_res and lo_res maps cannot "
                             "be aligned with shapes:\n%s\n%s" % (sh_hi,sh_lo))
        if diff > 0:
            crop_hi.append(diff // 2 )
            crop_lo.append(0)
        else:
            crop_lo.append(-diff // 2)
            crop_hi.append(0)

    if np.any(crop_lo):
        lo_res = Crop(lo_res, crop_lo, print_repr=True)
    if np.any(crop_hi):
        hi_res = Crop(hi_res, crop_hi, print_repr=True)

    out = Concat((lo_res, hi_res), axis='f', name='merge', print_repr=True)

    return out

UpConvMerge = ImageAlign

class Pool(Node):
    """
    Pooling layer.

    Reduces the count of training parameters by reducing the spatial size
    of its input by the factors given in ``pool_shape``.

    Pooling modes other than max-pooling can only be selected if cuDNN is
    available.

    Parameters
    ----------
    parent: Node
        The input node.
    pool_shape: tuple
        Tuple of pooling factors (per dimension) by which the input
        is downsampled.
    stride: tuple
        Stride sizes (per dimension).
    mfp: bool
        If max-fragment-pooling should be used.
    mode: str
        (only if cuDNN is available)
        Mode can be any of the modes supported by Theano's dnn_pool():
        ('max', 'average_inc_pad', 'average_exc_pad', 'sum').

        'max' (default): max-pooling
        'average' or 'average_inc_pad': average-pooling
        'sum': sum-pooling
    name: str
        Name of the pooling layer.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, parent, pool_shape, stride=None, mfp=False, mode='max',
                 name="pool", print_repr=True):
        super(Pool, self).__init__(parent, name, print_repr)

        if mfp and stride is not None:
            raise ValueError("Cannot use custom stride and MFP together")

        if stride is None:
            stride = pool_shape

        if mode == 'average':
            mode = 'average_inc_pad'  # Theano's internal name. 'average' is deprecated.

        self.pool_shape = pool_shape
        self.pool_stride = stride
        self.mfp = mfp
        self.mode = mode

        self.strides = parent.shape.strides
        self.mfp_offsets = parent.shape.mfp_offsets
        self.axis = parent.shape.tag2index('f') #retrieve feature shape's index
        self.axis_order = None

        spatial_axes = self.parent.shape.spatial_axes
        conv_dim = len(pool_shape)
        x_dim    = len(self.parent.shape)
        n_in = parent.shape['f'] #retrieve feature shape
        fail = False
        if conv_dim==1:
            if x_dim!=3 or spatial_axes!=[2]:
                fail = True
        elif conv_dim==2:
            if x_dim!=4 or spatial_axes!=[2,3]:
                fail = True

        elif conv_dim==3:
            if x_dim!=5:
                fail = True
            if spatial_axes==[2,3,4]:
                self.axis_order = 'dnn'
            elif spatial_axes==[1,3,4]:
                self.axis_order = 'theano'

            else:
                fail = True

        if fail:
            raise NotImplementedError("Cannot convolve non-standard shapes / axis orders. "
                                      "Implement reshaping before conv "
                                      "and re-reshaping afer!")

        self.spatial_axes = spatial_axes
        self.conv_dim = conv_dim


    def _make_output(self):
        """
        Computation of Theano output.
        """
        input_tensor = self.parent.output
        pattern  = ['x' for i in input_tensor.shape]
        pattern[self.axis] = 0

        if self.mfp:
            assert self.pool_stride == self.pool_shape
            if self.input_nodes[0].shape['b']!=1:
                raise ValueError("For MFP the batchsize of the raw image input must be 1.")
            lin_output, offsets_new, strides_new = computations.fragmentpool(input_tensor,
                                                                        self.pool_shape,
                                                                        self.mfp_offsets,
                                                                        self.strides,
                                                                        self.spatial_axes,
                                                                        mode=self.mode)
            self.mfp_offsets = offsets_new
            self.strides = strides_new
        else:
            lin_output = computations.pooling(input_tensor,self.pool_shape,
                                              self.spatial_axes, stride=self.pool_stride,
                                              mode=self.mode)
            self.strides = np.multiply(self.pool_stride, self.strides)

        self.output = lin_output


    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent.shape
        for j,(i,p,st) in enumerate(zip(self.spatial_axes , self.pool_shape, self.pool_stride)):
            tmp = sh[i] - p + st - 1
            s = tmp//st + 1
            if self.mfp:
                raise NotImplementedError("Check this first before use")
                if (tmp - p + 1)%st!=0:
                    raise ValueError("Cannot downsample spatial axis '%s' of length %i "
                                     "by factor %i with pool %i, and using MFP."\
                                     %(sh.tags[i], sh[i], st, p))
            else:
                if (tmp+1)%st!=0:
                    raise ValueError("Cannot downsample spatial axis '%s' of length %i "
                                     "by factor %i with pool %i."\
                                     %(sh.tags[i], sh[i], st, p ))
            sh = sh.updateshape(i, s)
            if sh.fov[j]>0:
                fov = sh.fov[j] + (p-1) * sh.strides[j]
            else:
                fov = -1
            sh = sh.updatefov(j, fov)

        if self.mfp:
            sh = sh.updatemfp_offsets(self.mfp_offsets)
            sh = sh.updateshape('b', np.prod(self.pool_shape), mode='mult')

        sh = sh.updatestrides(self.strides)
        self.shape = sh


class FaithlessMerge(Node):
    """
    FaithlessMerge node.

    Parameters
    ----------
    hard_features: Node
    easy_features: Node
    axis
    failing_prob: float
        The higher the more often merge is unreliable
    hardeasy_ratio: float
        The higher the more often the harder features fail instead of the easy ones
    name: str
            Name of the pooling layer.
        print_repr: bool
            Whether to print the node representation upon initialisation.
    """

    def __init__(self, hard_features, easy_features, axis='f', failing_prob=0.5,
                 hardeasy_ratio=0.8, name="faithless_merge", print_repr=True):
        parent_nodes = (hard_features, easy_features)
        super(FaithlessMerge, self).__init__(parent_nodes, name, print_repr)

        if isinstance(axis, str):
            self.axis = parent_nodes[0].shape.tag2index(axis)
        else:
            self.axis = axis

        failing_prob = VariableParam(value=failing_prob,
                                      name="failing_prob",
                                      dtype=floatX,
                                      apply_train=False)

        hardeasy_ratio = VariableParam(value=hardeasy_ratio,
                                     name="hardeasy_ratio",
                                     dtype=floatX,
                                     apply_train=False)


        self.params['failing_prob'] = failing_prob
        self.params['hardeasy_ratio'] = hardeasy_ratio
        self.failing_prob = failing_prob
        self.hardeasy_ratio = hardeasy_ratio


    def _make_output(self):
        """
        Computation of Theano output.
        """
        # It is assumed that all other dimensions are matching
        rng = T.shared_randomstreams.RandomStreams(int(time.time()))
        size = [1,] * self.parent[0].output.ndim
        axes = list(range(self.parent[0].output.ndim))

        not_failing = rng.binomial(size=size, n=1, p=self.failing_prob,
                                    dtype=theano.config.floatX)
        not_failing = T.addbroadcast(not_failing, *axes)
        hard_fails = rng.binomial(size=size, n=1, p=1-self.hardeasy_ratio,
                                    dtype=theano.config.floatX)
        hard_fails = T.addbroadcast(hard_fails, *axes)

        hard = self.parent[0].output * (1 - hard_fails * not_failing)
        easy = self.parent[1].output * (1 - (1 - hard_fails) * not_failing)

        self.output = T.concatenate([hard, easy], axis=self.axis)

    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        joint_axis_size = reduce(lambda x, y: x + y.shape[self.axis],
                                 self.parent, 0)
        # assuming all other dimensions are equal
        sh = self.parent[0].shape.updateshape(self.axis, joint_axis_size)
        self.shape = sh

    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.

        For this Node type this is hard-coded to 0.
        """
        self.computational_cost = 0


class GRU(NeuralLayer):
    """
    Gated Recurrent Unit Layer.

    Parameters
    ----------
    parent: Node
        The input node.
    memory_state: Node
        Memory node.
    n_f: int
        Number of features.
    activation_func: str
        Activation function name.
    flatten: bool
        (Unsupported).
    batch_normalisation: str or None
        Batch normalisation mode.
        Can be False (inactive), "train" or "fadeout".
    dropout_rate: float
        Dropout rate (probability that a node drops out in a training step).
    name: str
        Layer name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w: np.ndarray or T.TensorVariable
        (Unsupported).
        Weight matrix.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    b: np.ndarray or T.TensorVariable
        (Unsupported).
        Bias vector.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    gamma
        (For batch normalisation) Initializes gamma parameter.
    mean
        (For batch normalisation) Initializes mean parameter.
    std
        (For batch normalisation) Initializes std parameter.
    gradnet_mode
    """

    def __init__(self, parent, memory_state, n_f, activation_func='tanh',
                 flatten=False, batch_normalisation=False, dropout_rate=0,
                 name="gru", print_repr=True, w=None, b=None,
                 gamma=None, mean=None, std=None, gradnet_mode=None):
        parent_nodes = (parent, memory_state)
        super(GRU, self).__init__(parent_nodes, name, print_repr)

        self.n_f = n_f
        self.n_f_memory = memory_state.shape['f']
        self.activation_func = activation_func
        self.batch_normalisation = batch_normalisation
        self.gradnet_mode = gradnet_mode
        self.axis = parent.shape.tag2index('f') #retrieve feature shape's index
        self.spatial_axes = parent.shape.spatial_axes
        self.flatten = flatten

        if flatten:
            raise NotImplementedError("Flatten is not yet supported for GRU.")
            n_in = parent.shape.stripbatch_prod
        else:
            n_in = parent.shape['f']


        if self.n_f_memory != n_f:
            raise ValueError("n_f_memory != n_f not possible.")
        if parent.shape.hastag('r'):
            raise ValueError("Input must not have 'r' axis.")

        n_comb = self.n_f_memory + n_in
        if w != None or b != None:
             raise NotImplementedError("Initial weights are not yet supported for GRU.")

        w_sh = (n_comb, 3*n_f) #  [h_t-1, x] x [W_z/x, W_r/x, W_h/x]
        self._setup_params(w_sh, w, b, gamma, mean, std, dropout_rate)


    def _make_output(self):
        """
        Computation of Theano output.
        """
        parent = self.parent[0].output
        memory = self.parent[1].output
        pattern  = ['x' for i in parent.shape]
        pattern[self.axis] = 0
        broad_caster_shape = list(parent.shape)
        broad_caster_shape[self.axis] = self.n_f_memory
        broad_caster = T.ones(broad_caster_shape, dtype=memory.dtype)
        memory = memory * broad_caster
        input_tensor = T.concatenate([memory, parent] , axis=self.axis)

        activation_func = self.activation_func
        if activation_func.startswith("maxout"):
            r=int(activation_func.split(" ")[1])
            assert r>=2
            self.n_f /= r

        if activation_func=='prelu':
            b  = self.b[:-self.n_f,0].dimshuffle(pattern)
            b_h = self.b[-self.n_f:,0].dimshuffle(pattern)
            b1 = self.b[:-self.n_f,1].dimshuffle(pattern)
            b1_h = self.b[-self.n_f:,1].dimshuffle(pattern)
        else:
            b   = self.b[:-self.n_f].dimshuffle(pattern)
            b_h = self.b[-self.n_f:].dimshuffle(pattern)
            b1  = None
            b1_h = None

        lin_output = computations.dot(input_tensor, self.w[:, :-self.n_f], self.axis)

        if self.batch_normalisation in ['train', 'fadeout']:
            raise NotImplementedError("Batch normalisation not yet supported  for GRU.")
            mean = computations.apply_except_axis(
                lin_output,self.axis, T.mean).dimshuffle(pattern)
            std = computations.apply_except_axis(
                lin_output,self.axis, T.std).dimshuffle(pattern) + 1e-6
            gamma = self.gamma.dimshuffle(pattern)

            if self.batch_normalisation=='fadeout':
                logger.warning("Batch Normalisation mode 'fadeout' does not "
                               "work for less than 50%...")
                mean = self.gradnet_rate * mean
                std  = self.gradnet_rate * std + (1-self.gradnet_rate) * 1.0
                gamma = self.gradnet_rate * gamma

            self.mean.updates = (self.mean,
                                 0.9995 * self.mean + 0.0005 * T.extra_ops.squeeze(mean))
            self.std.updates  = (self.std,
                                 0.9995 * self.std + 0.0005 * T.extra_ops.squeeze(std))

        elif self.batch_normalisation=='predict':
            raise NotImplementedError("Batch normalisation not yet supported for GRU.")
            mean = self.mean.dimshuffle(pattern)
            std = self.std.dimshuffle(pattern)
            gamma = self.gamma.dimshuffle(pattern)
        else:
            mean = 0
            std = 1
            gamma = 1

        lin_output =  (gamma / std) * lin_output + b - (gamma * mean / std)
        act = computations.apply_activation(lin_output, 'sig', b1)
        slice_obj = [slice(None) for i in range(act.ndim)]
        slice_obj[self.axis] = slice(0, self.n_f)
        z = act[slice_obj]
        slice_obj[self.axis] = slice(self.n_f, None)
        r = act[slice_obj]
        gated_input = T.concatenate([r*memory, parent], axis=self.axis)
        h_tilde = computations.dot(gated_input, self.w[:, -self.n_f:], self.axis)
        h_tilde =  (gamma / std) *  h_tilde  + b_h - (gamma * mean / std)
        h_tilde = computations.apply_activation(h_tilde, activation_func, b1_h)
        act = (1 - z) * memory + z * h_tilde

        self._debug_outputs = [memory, act, z, r,]
        if self.dropout_rate:
            raise NotImplementedError("Dropout not yet supported for GRU.")
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p   = 1 - self.dropout_rate
            dropout_gate = rng.binomial(size=(self.n_f,), n=1, p=p,
                                        dtype=theano.config.floatX)
            dropout_gate *= 1.0 / p
            act =  act * dropout_gate.dimshuffle(('x', 0))

        self.output = act


    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent[0].shape
        if self.flatten:
            self.shape = TaggedShape((sh['b'], self.n_f), 'b,f')
        else:
            self.shape = sh.updateshape('f', self.n_f)


    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.
        """
        n = self.parent[0].shape.stripnone_prod
        self.computational_cost = 3 * n * self.n_f


class LSTM(NeuralLayer):
    """
    Long short term memory layer.

    Using an implementation without peepholes in f, i, o, i.e. weights
    cell state is not taken into account for weights. See
    http://colah.github.io/posts/2015-08-Understanding-LSTMs/.

    Parameters
    ----------
    parent: Node
        The input node.
    memory_states: Node
        Concatenated (initial) feed-back and cell state (one Node!).
    n_f: int
        Number of features.
    activation_func: str
        Activation function name.
    flatten
    batch_normalisation: str or None
        Batch normalisation mode.
        Can be False (inactive), "train" or "fadeout".
    dropout_rate: float
        Dropout rate (probability that a node drops out in a training step).
    name: str
        Layer name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    w: np.ndarray or T.TensorVariable
        Weight matrix.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    b: np.ndarray or T.TensorVariable
        Bias vector.
        If this is a np.ndarray, its values are used to initialise a
        shared variable for this layer.
        If it is a T.TensorVariable, it is directly used (weight sharing
        with the layer which this variable comes from).
    gamma
        (For batch normalisation) Initializes gamma parameter.
    mean
        (For batch normalisation) Initializes mean parameter.
    std
        (For batch normalisation) Initializes std parameter.
    gradnet_mode
    """

    def __init__(self, parent, memory_states, n_f, activation_func='tanh',
                 flatten=False, batch_normalisation=False, dropout_rate=0,
                 name="lstm", print_repr=True, w=None, b=None,
                 gamma=None, mean=None, std=None, gradnet_mode=None):
        parent_nodes = (parent, memory_states)
        super(LSTM, self).__init__(parent_nodes, name, print_repr)

        self.n_f = n_f
        self.n_f_memory = memory_states.shape['f']
        self.activation_func = activation_func
        self.batch_normalisation = batch_normalisation
        self.gradnet_mode = gradnet_mode
        self.axis = parent.shape.tag2index('f') #retrieve feature shape's index
        self.spatial_axes = parent.shape.spatial_axes
        self.flatten = flatten

        if flatten:
            raise NotImplementedError("Flatten is not yet supported for LSTM.")
        else:
            n_in = parent.shape['f']

        n_comb = n_f + n_in
        if w != None or b != None:
             raise NotImplementedError("Initial weights are not yet supported for LSTM.")

        if self.n_f_memory != 2*n_f:
            raise ValueError("n_f of memory_states must be 2*n_f.")
        if parent.shape.hastag('r'):
            raise ValueError("Input must not have 'r' axis.")

        w_sh = (n_comb, 4*n_f) # f, i, o, C
        self._setup_params(w_sh, w, b, gamma, mean, std, dropout_rate)


    def _make_output(self):
        """
        Computation of Theano output.
        """
        parent = self.parent[0].output
        memory = self.parent[1].output

        broad_caster_shape = list(parent.shape)
        broad_caster_shape[self.axis] = self.n_f_memory
        broad_caster = T.ones(broad_caster_shape, dtype=memory.dtype)
        memory = memory * broad_caster

        slice_obj = [slice(None) for i in range(len(self.parent[1].shape))]
        slice_obj[self.parent[1].shape.tag2index('f')] = slice(self.n_f)
        feed_back = memory[slice_obj]
        slice_obj[self.parent[1].shape.tag2index('f')] = slice(self.n_f, None)
        cell_state = memory[slice_obj]


        input_tensor = T.concatenate([feed_back, parent] ,
                                     axis=self.axis) #h, x
        pattern = ['x' for i in input_tensor.shape]
        pattern[self.axis] = 0
        activation_func = self.activation_func

        if activation_func.startswith("maxout"):
            r=int(activation_func.split(" ")[1])
            assert r>=2
            self.n_f /= r

        if activation_func=='prelu':
            b  = self.b[:, 0].dimshuffle(pattern)
            b1 = self.b[:, 1]
            b1_f = b1[:self.n_f].dimshuffle(pattern)
            b1_i = b1[self.n_f:2*self.n_f].dimshuffle(pattern)
            b1_o = b1[2*self.n_f:3*self.n_f].dimshuffle(pattern)
            b1_c = b1[3*self.n_f:].dimshuffle(pattern)
        else:
            b   = self.b.dimshuffle(pattern)
            b1_f = None
            b1_i = None
            b1_o = None
            b1_c = None

        lin_output = computations.dot(input_tensor, self.w, self.axis)

        if self.batch_normalisation in ['train', 'fadeout']:
            raise NotImplementedError("Batch normalisation not yet supported for LSTM.")
            mean = computations.apply_except_axis(
                lin_output,self.axis, T.mean).dimshuffle(pattern)
            std = computations.apply_except_axis(
                lin_output,self.axis, T.std).dimshuffle(pattern) + 1e-6
            gamma = self.gamma.dimshuffle(pattern)

            if self.batch_normalisation=='fadeout':
                logger.warning("Batch Normalisation mode 'fadeout' does not "
                               "work for less than 50%...")
                mean = self.gradnet_rate * mean
                std  = self.gradnet_rate * std + (1-self.gradnet_rate) * 1.0
                gamma = self.gradnet_rate * gamma

            self.mean.updates = (self.mean,
                                 0.9995 * self.mean + 0.0005 * T.extra_ops.squeeze(mean))
            self.std.updates  = (self.std,
                                 0.9995 * self.std + 0.0005 * T.extra_ops.squeeze(std))

        elif self.batch_normalisation=='predict':
            raise NotImplementedError("Batch normalisation not yet supported for LSTM.")
            mean = self.mean.dimshuffle(pattern)
            std = self.std.dimshuffle(pattern)
            gamma = self.gamma.dimshuffle(pattern)
        else:
            mean = 0
            std = 1
            gamma = 1

        lin_output =  (gamma / std) * lin_output + b - (gamma * mean / std)
        slice_obj = [slice(None) for i in range(lin_output.ndim)]
        slice_obj[self.axis] = slice(self.n_f)
        f = computations.apply_activation(lin_output[slice_obj], 'sig', b1_f)
        slice_obj[self.axis] = slice(self.n_f, 2*self.n_f)
        i = computations.apply_activation(lin_output[slice_obj], 'sig', b1_i)
        slice_obj[self.axis] = slice(2*self.n_f, 3*self.n_f)
        o = computations.apply_activation(lin_output[slice_obj], 'sig', b1_o)
        slice_obj[self.axis] = slice(3*self.n_f, 4*self.n_f)
        c_tilde = computations.apply_activation(lin_output[slice_obj], activation_func, b1_c)
        cell_out = f * cell_state + i * c_tilde
        lin_output = o * computations.apply_activation(cell_out, activation_func, None)

        if self.dropout_rate:
            raise NotImplementedError("Dropout not yet supported for LSTM.")
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p   = 1 - self.dropout_rate
            dropout_gate = rng.binomial(size=(self.n_f,), n=1, p=p,
                                        dtype=theano.config.floatX)
            dropout_gate *= 1.0 / p
            lin_output =  lin_output * dropout_gate.dimshuffle(('x', 0))

        self.output = T.concatenate([lin_output, cell_out], axis=self.axis)


    def _calc_shape(self):
        """
        Calculate and set self.shape.
        """
        sh = self.parent[0].shape
        if self.flatten:
            self.shape = TaggedShape((sh['b'], 2*self.n_f), 'b,f')
        else:
            self.shape = sh.updateshape('f',2* self.n_f)


    def _calc_comp_cost(self):
        """
        Calculate and set self.computational_cost.
        """
        n = self.parent[0].shape.stripnone_prod
        self.computational_cost = 4 * n * self.n_f


class LRN(Node):
    """
    LRN (Local Response Normalization) layer.

    Parameters
    ----------
    parent: Node
        The input node.
    filter_shape: tuple
    mode: str
        Can be "spatial" or "channel".
    alpha: float
    k: float
    beta: float
    name: str
        Node name.
    print_repr: bool
        Whether to print the node representation upon initialisation.
    """

    def __init__(self, parent, filter_shape, mode='spatial',  alpha=0.0001,
                 k=1, beta=0.75, name="LRN", print_repr=True):
        super(LRN, self).__init__(parent, name, print_repr)

        self.mode = mode
        self.filter_shape = filter_shape
        self.axis = parent.shape.tag2index('f')  # retrieve feature shape's index
        if mode=='spatial':
            self.axis_order = None
            self.spatial_axes = self.parent.shape.spatial_axes
            conv_dim = len(self.spatial_axes)
            x_dim    = len(self.parent.shape)
            if len(self.spatial_axes)!=len(filter_shape):
                raise ValueError("The filter_shape dimensionality (%i) and the number "
                                 "of spatial dimensions in the input (%i)differ! "
                                 "Use filter size 1 on axes which should not be "
                                 "averaged."\
                                 %(len(filter_shape), conv_dim, ))

            n_in = parent.shape['f'] #retrieve feature shape
            fail = False
            if conv_dim==1:
                if x_dim!=3 or self.spatial_axes!=[2]:
                    fail = True
                w_sh = [n_in, n_in] + list(filter_shape)

            elif conv_dim==2:
                if x_dim!=4 or self.spatial_axes!=[2,3]:
                    fail = True
                w_sh = [n_in, n_in] + list(filter_shape)

            elif conv_dim==3:
                if x_dim!=5:
                    fail = True
                if self.spatial_axes==[2,3,4]:
                    self.axis_order = 'dnn'
                    w_sh = [n_in, n_in] + list(filter_shape)
                elif self.spatial_axes==[1,3,4]:
                    self.axis_order = 'theano'
                    w_sh = [n_in, filter_shape[0], n_in] + list(filter_shape[1:])
                else:
                    fail = True

            if fail:
                raise NotImplementedError("Cannot convolve non-standard shapes / axis orders. "
                                          "Implement reshaping before conv"
                                          "and re-reshaping afer!")

            self.conv_dim = conv_dim
            self.w_sh = w_sh
            value = np.zeros(w_sh, dtype=floatX)
            val = 1.0 / np.product(filter_shape)
            for i in range(n_in):
                value[i,i] = val

            self.average_filter = ConstantParam(value, '<%s_filter%s>'%(self.name, tuple(w_sh)))
            self.params['average_filter'] = self.average_filter
        elif mode=='channel':
            assert isinstance(filter_shape, int)
            assert filter_shape%2==1
        else:
            raise ValueError("Unknow mode %s"%mode)

        self.alpha = VariableParam(value=alpha, name="alpha",
                                   dtype=floatX, apply_train=False)

        self.beta = VariableParam(value=beta,name="beta",
                                  dtype=floatX, apply_train=False)

        self.k = VariableParam(value=k,name="k",
                               dtype=floatX, apply_train=False)

        self.params['alpha'] = self.alpha
        self.params['beta'] = self.beta
        self.params['k'] = self.k


    def _make_output(self):
        """
        Computation of Theano output.
        """
        input_tensor = self.parent.output
        input_shape  = list(self.parent.shape)

        if self.mode=='spatial':
            mean_square = computations.conv(T.square(input_tensor), self.average_filter,
                                           self.axis_order, border_mode='same',
                                           x_shape=input_shape, w_shape=self.w_sh)
        else:
            n_f = input_shape[self.axis]
            in_square = T.square(input_tensor)
            half_n = self.filter_shape // 2
            new_sh = list(input_tensor.shape)
            new_sh[self.axis] += 2 * half_n
            in_square_ext = T.zeros(new_sh, floatX)
            slicer = [slice(None)] * input_tensor.ndim
            slicer[self.axis] = slice(half_n,half_n+n_f)
            in_square_ext = T.set_subtensor(in_square_ext[slicer], in_square)
            # pad left
            slicer[self.axis] = slice(0, half_n)
            pad_slicer = [slice(None)] * input_tensor.ndim
            pad_slicer[self.axis] = slice(0, 1)
            in_square_ext = T.set_subtensor(in_square_ext[slicer], in_square[pad_slicer])
            # pad right
            slicer[self.axis] = slice(half_n+n_f, 2*half_n+n_f)
            pad_slicer[self.axis] = slice(n_f-1,n_f)
            in_square_ext = T.set_subtensor(in_square_ext[slicer], in_square[pad_slicer])

            mean_square = 0
            for i in range(self.filter_shape):
                slicer[self.axis] = slice(i,i+n_f)
                mean_square += in_square_ext[slicer]

            mean_square /= self.filter_shape


        divisor = T.power(self.k + self.alpha * mean_square, self.beta)
        self.output = input_tensor / divisor
        self._debug_outputs = [mean_square, divisor]



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved

from __future__ import absolute_import

from .utils_basic import *
from .cnncalculator import *



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Martin Drawitsch, Marius Killinger
# All rights reserved

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


__all__ = ['warp_slice', 'get_tracing_slice', 'WarpingOOBError',
           'Transform', 'trafo_from_array', 'get_warped_slice']


import itertools
import logging
from functools import reduce

import numpy as np
import numba

from .. import utils

logger = logging.getLogger('elektronn2log')
inspection_logger = logging.getLogger('elektronn2log-inspection')


# @numba.jit(nopython=True, cache=True)
# def map_coordinates_nearest(src, coords, lo, dest):
#     sh = coords.shape
#     for z in np.arange(sh[0]):
#         for x in np.arange(sh[1]):
#             for y in np.arange(sh[2]):
#                 u = np.int32(np.round(coords[z,x,y,0] - lo[0]))
#                 v = np.int32(np.round(coords[z,x,y,1] - lo[1]))
#                 w = np.int32(np.round(coords[z,x,y,2] - lo[2]))
#                 or
#                 u = coords[z, x, y, 0] - lo[0]
#                 v = coords[z, x, y, 1] - lo[1]
#                 w = coords[z, x, y, 2] - lo[2]
#                 dest[z,x,y] = src[u,v,w]

@numba.guvectorize(['void(float32[:,:,:], float32[:], float32[:], float32[:,],)'],
              '(x,y,z),(i),(i)->()', target='parallel', nopython=True)
def map_coordinates_nearest(src, coords, lo, dest):
    u = np.int32(np.round(coords[0] - lo[0]))
    v = np.int32(np.round(coords[1] - lo[1]))
    w = np.int32(np.round(coords[2] - lo[2]))
    dest[0] = src[u,v,w]

@numba.guvectorize(['void(float32[:,:,:], float32[:], float32[:], float32[:,],)'],
              '(x,y,z),(i),(i)->()', target='parallel', nopython=True)
def map_coordinates_linear(src, coords, lo, dest):
    u = coords[0] - lo[0]
    v = coords[1] - lo[1]
    w = coords[2] - lo[2]
    u0 = np.int32(u)
    u1 = u0 + 1
    du = u - u0
    v0 = np.int32(v)
    v1 = v0 + 1
    dv = v - v0
    w0 = np.int32(w)
    w1 = w0 + 1
    dw = w - w0
    val = src[u0, v0, w0] * (1-du) * (1-dv) * (1-dw) +\
          src[u1, v0, w0] * du * (1-dv) * (1-dw) +\
          src[u0, v1, w0] * (1-du) * dv * (1-dw) +\
          src[u0, v0, w1] * (1-du) * (1-dv) * dw +\
          src[u1, v0, w1] * du * (1-dv) * dw +\
          src[u0, v1, w1] * (1-du) * dv * dw +\
          src[u1, v1, w0] * du * dv * (1-dw) +\
          src[u1, v1, w1] * du * dv * dw
    dest[0] = val

@numba.jit(nopython=True, cache=True)
def map_coordinates_max_kernel(src, coords, lo, k, dest):
    k = np.float32(k)
    kz = min(0.5, k/2)
    sh = coords.shape
    sh_src = src.shape
    for z in np.arange(sh[0]):
        for x in np.arange(sh[1]):
            for y in np.arange(sh[2]):
                u = coords[z,x,y,0] - lo[0]
                v = coords[z,x,y,1] - lo[1]
                w = coords[z,x,y,2] - lo[2]

                u0 = np.int32(np.round(max(0, min(sh_src[0], u - kz))))
                u1 = np.int32(np.round(max(0, min(sh_src[0], u + kz))))
                v0 = np.int32(np.round(max(0, min(sh_src[1], v - k))))
                v1 = np.int32(np.round(max(0, min(sh_src[1], v + k))))
                w0 = np.int32(np.round(max(0, min(sh_src[2], w - k))))
                w1 = np.int32(np.round(max(0, min(sh_src[2], w + k))))

                val = src[u0:u1, v0:v1, w0:w1].max()

                dest[z, x, y] = val


def identity():
    return np.eye(4, dtype=np.float32)

def translate(dz, dy, dx):
    return np.array([
        [1.0, 0.0, 0.0,  dz],
        [0.0, 1.0, 0.0,  dy],
        [0.0, 0.0, 1.0,  dx],
        [0.0, 0.0, 0.0, 1.0]
    ], dtype=np.float32)

def rotate_z(a):
    return np.array([
        [1.0, 0.0,    0.0,     0.0],
        [0.0, np.cos(a), -np.sin(a), 0.0],
        [0.0, np.sin(a), np.cos(a),  0.0],
        [0.0, 0.0,    0.0,     1.0]
    ], dtype=np.float32)

def rotate_y(a):
    return np.array([
        [np.cos(a), -np.sin(a), 0.0, 0.0],
        [np.sin(a),  np.cos(a), 0.0, 0.0],
        [0.0,        0.0, 1.0, 0.0],
        [0.0,        0.0, 0.0, 1.0]
    ], dtype=np.float32)

def rotate_x(a):
    return np.array([
        [np.cos(a),  0.0, np.sin(a), 0.0],
        [0.0,     1.0, 0.0,    0.0],
        [-np.sin(a), 0.0, np.cos(a), 0.0],
        [0.0,     0.0, 0.0,    1.0]
    ], dtype=np.float32)

def scale_inv(mz, my, mx):
    return np.array([
        [1/mz,  0.0,    0.0,  0.0],
        [0.0,   1/my,   0.0,  0.0],
        [0.0,   0.0,    1/mx, 0.0],
        [0.0,   0.0,    0.0,  1.0]
    ], dtype=np.float32)

def scale(mz, my, mx):
    return np.array([
        [mz,  0.0, 0.0, 0.0],
        [0.0, my,  0.0, 0.0],
        [0.0, 0.0, mx,  0.0],
        [0.0, 0.0, 0.0, 1.0]
    ], dtype=np.float32)

def chain_matrices(mat_list):
    return reduce(np.dot, mat_list, identity())

def get_euler_angles(direc, gamma):
    """
    tracing_dir (z, x, y) normalised
    angle3 rotation around z in dest frame
    phi is the rotation about the 1-2 axis
    theta is the rotation about the 0'-1' axis
    """
    assert abs(np.linalg.norm(direc) - 1) < 1e-3
    phi = np.arctan2(direc[2], direc[1])
    theta = np.arccos(direc[0])
    return phi, theta, gamma

def get_rotmat_from_direc(direc, gamma=None, rng=None):
    if gamma is None:
        gamma=0.0
    elif gamma=='rand':
        if rng is None:
            gamma = np.random.rand() * 2 * np.pi
        else:
            gamma = rng.rand() * 2 * np.pi

    phi, theta, gamma = get_euler_angles(direc, gamma)
    R1 = rotate_z(-phi)
    R2 = rotate_y(-theta)
    R3 = rotate_z(gamma)
    R = chain_matrices([R3, R2, R1])
    return R

def get_random_rotmat(lock_z=False, amount=1.0, rng=None):
    rng = np.random.RandomState() if rng is None else rng

    gamma = rng.rand() * 2 * np.pi * amount
    if lock_z:
        return rotate_z(gamma)

    phi = rng.rand() * 2 * np.pi * amount
    theta = np.arcsin(rng.rand()) * amount

    R1 = rotate_z(-phi)
    R2 = rotate_y(-theta)
    R3 = rotate_z(gamma)
    R = chain_matrices([R3, R2, R1])
    return R

def get_random_flipmat(no_x_flip=False, rng=None):
    rng = np.random.RandomState() if rng is None else rng
    F = np.eye(4, dtype=np.float32)
    flips = rng.binomial(1, 0.5, 4) * 2 - 1
    flips[3] = 1 # don't flip homogeneous dimension
    if no_x_flip:
        flips[2] = 1

    np.fill_diagonal(F, flips)
    return F

def get_random_swapmat(lock_z=False, rng=None):
    rng = np.random.RandomState() if rng is None else rng
    S = np.eye(4, dtype=np.float32)
    if lock_z:
        swaps = [[0, 1, 2, 3],
                 [0, 2, 1, 3]]
    else:
        swaps = [[0, 1, 2, 3],
                 [0, 2, 1, 3],
                 [1, 0, 2, 3],
                 [1, 2, 0, 3],
                 [2, 0, 1, 3],
                 [2, 1, 0, 3]]

    i = rng.randint(0, len(swaps))
    S = S[swaps[i]]
    return S

def get_random_warpmat(lock_z=False, perspective=False, amount=1.0, rng=None):
    W = np.eye(4, dtype=np.float32)
    amount *= 0.1
    perturb = np.random.uniform(-amount, amount, (4, 4))
    perturb[3,3] = 0
    if lock_z:
        perturb[0] = 0
        perturb[:,0] = 0
    if not perspective:
        perturb[3] = 0

    perturb[3,:3] *= 0.05 # perspective parameters need to be very small
    np.clip(perturb[3,:3], -3e-3, 3e-3, out=perturb[3,:3])

    return W + perturb

@utils.cache()
def make_dest_coords(sh):
    """
    Make coordinate list for destiantion array of shape sh
    """
    zz,xx,yy = np.mgrid[0:sh[0], 0:sh[1], 0:sh[2]]
    hh = np.ones(sh, dtype=np.int)
    coords = np.concatenate([zz[...,None], xx[...,None],
                             yy[...,None], hh[...,None]], axis=-1)
    return coords.astype(np.float32)

def make_dest_corners(sh):
    """
    Make coordinate list of the corners of destination array of shape sh
    """
    corners = list(itertools.product(*([0,1],)*3))
    sh = np.subtract(sh, 1) # 0-based indices
    corners = np.multiply(sh, corners)
    corners = np.hstack((corners, np.ones((8,1)))) # homogeneous coords
    return corners


class WarpingOOBError(ValueError):
    def __init__(self, *args, **kwargs):
        super(WarpingOOBError, self).__init__( *args, **kwargs)


class Transform(object):
    def __init__(self, M, position_l=None, aniso_factor=2):
        self.M = M
        self.M_inv = np.linalg.inv(M.astype(np.float64)).astype(np.float32) # stability...
        self.position_l = position_l
        self.aniso_factor = aniso_factor
        self.is_projective = not np.allclose(M[3,:3], 0.0)

    @property
    def M_lin(self):
        if self.is_projective:
            raise ValueError("This transform requires homogeneous coordinates")
        else:
            return self.M[:3,:3]

    @property
    def M_lin_inv(self):
        if self.is_projective:
            raise ValueError("This transform requires homogeneous coordinates")
        else:
            return self.M_inv[:3, :3]

    def to_array(self):
        return np.hstack([self.M.ravel(), self.position_l, self.aniso_factor])

    def lab_coord2cnn_coord(self, vec_l):
        assert not self.is_projective
        if vec_l.ndim==1:
            vec_c = np.dot(self.M_lin, vec_l)  # rotation
        else:
            # assume vec_l.shape=(n,3)
            assert vec_l.shape[1]==3
            vec_c = np.dot(vec_l, self.M_lin.T)  # rotation
        return vec_c


    def cnn_coord2lab_coord(self, vec_c, add_offset_l=False):
        assert not self.is_projective
        if vec_c.ndim==1:
            vec_l = np.dot(self.M_lin_inv, vec_c)  # rotation
            if add_offset_l:
                vec_l += self.position_l
        else:
            # assume vec_l.shape=(n,3)
            assert vec_c.shape[1]==3
            vec_l = np.dot(vec_c, self.M_lin_inv.T)  # rotation
            if add_offset_l:
                vec_l += self.position_l[None,:]
        return vec_l


    def cnn_pred2lab_position(self, prediction_c):
        assert not self.is_projective
        tracin_direc_l = self.cnn_coord2lab_coord(prediction_c, add_offset_l=False)
        new_position_l = tracin_direc_l + self.position_l
        tracing_direc_il = tracin_direc_l * [self.aniso_factor,1,1]
        assert np.linalg.norm(tracing_direc_il) > 0 # normalise
        tracing_direc_il /= np.linalg.norm(tracing_direc_il)
        return new_position_l, tracing_direc_il

def trafo_from_array(a):
    M = a[:16].reshape((4,4))
    offset_l = a[16:19]
    aniso_factor = a[19]
    return Transform(M, offset_l, aniso_factor)


def warp_slice(img, ps, M, target=None, target_ps=None,
               target_vec_ix=None, target_discrete_ix=None,
               last_ch_max_interp=False, ksize=0.5):
    """
    Cuts a warped slice out of the input image and out of the target image.
    Warping is applied by multiplying the original source coordinates with
    the inverse of the homogenous (forward) transformation matrix ``M``.
    
    "Source coordinates" (``src_coords``) signify the coordinates of voxels in
    ``img`` and ``target`` that are used to compose their respective warped
    versions. The idea here is that not the images themselves, but the
    coordinates from where they are read are warped. his allows for much higher
    efficiency for large image volumes because we don't have to calculate the
    expensive warping transform for the whole image, but only for the voxels
    that we eventually want to use for the new warped image.
    The transformed coordinates usually don't align to the discrete
    voxel grids of the original images (meaning they are not integers), so the
    new voxel values are obtained by linear interpolation.
    
    Parameters
    ----------
    img: np.ndarray
        Image array in shape ``(f, z, x, y)``
    ps: tuple
        (spatial only) Patch size ``(z, x, y)``
        (spatial shape of the neural network's input node)
    M: np.ndarray
        Forward warping tansformation matrix (4x4).
        Must contain translations in source and target array.
    target: np.ndarray or None
        Optional target array to be extracted in the same way.
    target_ps: tuple
        Patch size for the ``target`` array.
    target_vec_ix: list
        List of triples that denote vector value parts in the target array.
        E.g. [(0,1,2), (4,5,6)] denotes two vector fields, separated by a
        scalar field in channel 3.
    last_ch_max_interp: bool
    ksize: float

    Returns
    -------
    img_new: np.ndarray
        Warped input image slice
    target_new: np.ndarray or None
        Warped target image slice
        or ``None``, if ``target is None``.
    """

    #T = utils.Timer(silent_all=True)
    ps = tuple(ps)
    if len(img.shape)==3: # For single knossos_array
        n_f = 1
        sh = img.shape
    elif len(img.shape)==4:
        n_f = img.shape[0]
        sh = img.shape[1:]
    else:
        raise ValueError('img wrong dim/shape')

    M_inv = np.linalg.inv(M.astype(np.float64)).astype(np.float32) # stability...
    dest_corners = make_dest_corners(ps)
    src_corners = np.dot(M_inv, dest_corners.T).T
    if np.any(M[3,:3] != 0): # homogeneous divide
        src_corners /= src_corners[:,3][:,None]

    # check corners
    src_corners = src_corners[:,:3]
    lo = np.min(np.floor(src_corners), 0).astype(np.int)
    hi = np.max(np.ceil(src_corners + 1), 0).astype(np.int) # add 1 because linear interp
    if np.any(lo < 0) or np.any(hi >= sh):
        raise WarpingOOBError("Out of bounds")
    #T.check("corners")
    # compute/transform dense coords
    dest_coords = make_dest_coords(ps)
    src_coords = np.tensordot(dest_coords, M_inv, axes=[[-1],[1]])
    if np.any(M[3,:3] != 0): # homogeneous divide
        src_coords /= src_coords[...,3][...,None]
    #T.check("tensordot")
    # cut patch
    src_coords = src_coords[...,:3]
    img_cut = img[:, lo[0]:hi[0]+1, #add 1 to include this coordinate!
                     lo[1]:hi[1]+1,
                     lo[2]:hi[2]+1,]
    img_cut = np.ascontiguousarray(img_cut, dtype=np.float32)
    #T.check("cut")
    img_new = np.zeros((n_f,)+ps, dtype=np.float32)
    #T.check("alloc")
    lo = lo.astype(np.float32)
    for k in range(n_f):
        if (ksize>0.5) and last_ch_max_interp and (k==n_f-1):
            map_coordinates_max_kernel(img_cut[k], src_coords, lo, ksize, img_new[k])
        else:
            map_coordinates_linear(img_cut[k], src_coords, lo, img_new[k])
    #T.check("map img")

    if np.isnan(img_new).sum():
        logger.warning("img_new has {} NaN values".format(np.isnan(img_new).sum()))

    if target is not None:
        target_ps = tuple(target_ps)
        n_f_t = target.shape[0]

        off = np.subtract(sh, target.shape[1:])
        if np.any(np.mod(off, 2)):
            raise ValueError("targets must be centered w.r.t. images")
        off //= 2

        off_ps = np.subtract(ps, target_ps)
        if np.any(np.mod(off_ps, 2)):
            raise ValueError("targets must be centered w.r.t. images")
        off_ps //= 2

        src_coords_target = src_coords[off_ps[0]:off_ps[0]+target_ps[0],
                                       off_ps[1]:off_ps[1]+target_ps[1],
                                       off_ps[2]:off_ps[2]+target_ps[2]]
        # shift coords to be w.r.t to origin of target array
        lo_targ = np.floor(src_coords_target.min(2).min(1).min(0) - off).astype(np.int)
        # add 1 because linear interp
        hi_targ = np.ceil(src_coords_target.max(2).max(1).max(0) - off + 1).astype(np.int)
        if np.any(lo_targ < 0) or np.any(hi_targ >= target.shape[1:]):
             raise WarpingOOBError("Out of bounds for target")
        #T.check("target_check")
        target_cut = target[:, lo_targ[0]:hi_targ[0]+1, #add 1 to include this coordinate!
                               lo_targ[1]:hi_targ[1]+1,
                               lo_targ[2]:hi_targ[2]+1]

        target_cut = np.ascontiguousarray(target_cut, dtype=np.float32)
        src_coords_target = np.ascontiguousarray(src_coords_target, dtype=np.float32)
        #T.check("target_cut")
        target_new = np.zeros((n_f_t,)+target_ps, dtype=np.float32)
        #T.check("target_alloc")
        lo_targ = (lo_targ + off).astype(np.float32)
        if target_discrete_ix is None:
            target_discrete_ix = [True for i in range(n_f_t)]
        else:
            target_discrete_ix = [i in target_discrete_ix for i in range(n_f_t)]

        for k, discr in enumerate(target_discrete_ix):
            if discr:
                map_coordinates_nearest(target_cut[k], src_coords_target, lo_targ, target_new[k])
            else:
                map_coordinates_linear(target_cut[k], src_coords_target, lo_targ, target_new[k])

        #T.check("target_map")
        if target_vec_ix is not None: # Vectors must be transformed again
            assert np.allclose(M[3,:3], 0.0) # no projective transform
            M_lin = M[:3,:3]
            for ix in target_vec_ix:
                assert len(ix)==3
                target_new[ix] = np.tensordot(M_lin, target_new[ix], axes=[[1],[0]])
        #T.check("target_vec")
    else:
        target_new =  None
    #T.summary()
    return img_new, target_new

def get_tracing_slice(img, ps, pos, z_shift=0, aniso_factor=2,
                      sample_aniso=True, gamma=0, scale_factor=1.0, direction_iso=None,
                      target=None, target_ps=None, target_vec_ix=None,
                      target_discrete_ix=None, rng=None, last_ch_max_interp=False):

    # positive z_shift --> see more slices in positive z-direction w.r.t pos
    # scale_factor > 1 zooms into image / magifies
    rng = np.random.RandomState() if rng is None else rng
    dest_center = np.array(ps, dtype=np.float)/2
    dest_center[0] -= z_shift
    R = get_rotmat_from_direc(direction_iso, gamma, rng)
    T_src = translate(-pos[0], -pos[1], -pos[2])
    S_src = scale(aniso_factor, 1, 1)
    S_zoom = scale(scale_factor, scale_factor, scale_factor)

    if sample_aniso:
        S_dest = scale_inv(aniso_factor, 1, 1)
    else:
        S_dest = identity()
    T_dest = translate(dest_center[0], dest_center[1], dest_center[2])

    M = chain_matrices([T_dest, S_zoom, S_dest, R, S_src, T_src])
    ksize = min(0.5, 0.5/scale_factor)
    img_new, target_new = warp_slice(img, ps, M,
                     target=target,
                     target_ps=target_ps,
                     target_vec_ix=target_vec_ix,
                     target_discrete_ix=target_discrete_ix,
                     last_ch_max_interp=last_ch_max_interp,
                     ksize=ksize)

    return img_new, target_new, M


def get_warped_slice(img, ps, aniso_factor=2, sample_aniso=True,
                    warp_amount=1.0, lock_z=True, no_x_flip=False, perspective=False,
                    target=None, target_ps=None, target_vec_ix=None,
                    target_discrete_ix=None, rng=None):
    """
    (Wraps :py:meth:`elektronn2.data.transformations.warp_slice()`)
    
    Generates the warping transformation parameters and composes them into a
    single 4D homogenous transformation matrix ``M``.
    Then this transformation is applied to ``img`` and ``target`` in the
    ``warp_slice()`` function and the transformed input and target image are
    returned.
    
    Parameters
    ----------
    img: np.array
        Input image
    ps: np.array
        Patch size (spatial shape of the neural network's input node)
    aniso_factor: float
        Anisotropy factor that determines an additional scaling in ``z``
        direction.
    sample_aniso: bool
        Scale coordinates by ``1 / aniso_factor`` while warping.
    warp_amount: float
        Strength of the random warping transformation. A lower ``warp_amount``
        will lead to less distorted images.
    lock_z: bool
        Exclude ``z`` coordinates from the random warping transformations.
    no_x_flip: bool
        Don't flip ``x`` axis during random warping.
    perspective: bool
        Apply perspective transformations (in addition to affine ones).
    target: np.array
        Target image
    target_ps: np.array
        Target patch size
    target_vec_ix
    target_discrete_ix
    rng: np.random.mtrand.RandomState
        Random number generator state (obtainable by
        ``np.random.RandomState()``). Passing a known state makes the random
        transformations reproducible.

    Returns
    -------
    img_new: np.ndarray
        (Warped) input image slice
    target_new: np.ndarray
        (Warped) target slice
    """

    rng = np.random.RandomState() if rng is None else rng


    strip_2d = False
    if len(ps)==2:
        strip_2d = True
        ps = np.array([1]+list(ps))
        if target is not None:
            target_ps = np.array([1]+list(target_ps))


    dest_center = np.array(ps, dtype=np.float) / 2
    src_remainder =  np.array(np.mod(ps, 2), dtype=np.float) / 2

    if target_ps is not None:
        t_center = np.array(target_ps, dtype=np.float) / 2
        off = np.subtract(img.shape[1:], target.shape[1:])
        off //= 2
        lo_pos = np.maximum(dest_center, t_center+off)
        hi_pos = np.minimum(img.shape[1:] - dest_center, target.shape[1:] - t_center + off)
    else:
        lo_pos = dest_center
        hi_pos = img.shape[1:] - dest_center

    z = rng.randint(lo_pos[0], hi_pos[0]) + src_remainder[0]
    y = rng.randint(lo_pos[1], hi_pos[1]) + src_remainder[1]
    x = rng.randint(lo_pos[2], hi_pos[2]) + src_remainder[2]

    F = get_random_flipmat(no_x_flip, rng)
    if no_x_flip:
        S = np.eye(4, dtype=np.float32)
    else:
        S = get_random_swapmat(lock_z, rng)

    if np.isclose(warp_amount, 0):
        R = np.eye(4, dtype=np.float32)
        W = np.eye(4, dtype=np.float32)
    else:
        R = get_random_rotmat(lock_z, warp_amount, rng)
        W = get_random_warpmat(lock_z, perspective, warp_amount, rng)

    T_src = translate(-z, -y, -x)
    S_src = scale(aniso_factor, 1, 1)

    if sample_aniso:
        S_dest = scale(1.0 / aniso_factor, 1, 1)
    else:
        S_dest = identity()
    T_dest = translate(dest_center[0], dest_center[1], dest_center[2])

    M = chain_matrices([T_dest, S_dest, R, W, F, S, S_src, T_src])

    img_new, target_new = warp_slice(img, ps, M,
                                     target=target,
                                     target_ps=target_ps,
                                     target_vec_ix=target_vec_ix,
                                     target_discrete_ix=target_discrete_ix)

    if strip_2d:
        img_new = img_new[:,0]
        if target is not None:
            target_new = target_new[:,0]

    return img_new, target_new





if __name__ =='__main__':
#    pos = [120,175,175]
#    import time
#    from elektronn2 import utils
#    from elektronn2.utils.plotting import scroll_plot
#
#    raw = np.random.rand(250,500,500).astype(np.float32)
#    if True: # insert stripes
#        raw[::10,:,:] = 0
#        raw[:,::20,:] = 0
#        raw[:,:,::20] = 0
#
#    img = raw[None]
#    R = chain_matrices([rotate_y(0.2), rotate_z(0.1)])
#    ps = [46,171,171]
#    z_shift = 8
#    aniso_factor = 2
#    sample_aniso=True
#
#    dest_center = np.array(ps, dtype=np.float32)/2
#    dest_center[0] -= z_shift # sign?
#    T_src = translate(-pos[0], -pos[1], -pos[2])
#    S_src = scale(aniso_factor, 1, 1)
#    if sample_aniso:
#        S_dest = scale(1.0/aniso_factor, 1, 1)
#    else:
#        S_dest = identity()
#    T_dest = translate(dest_center[0], dest_center[1], dest_center[2])
#
#    M = chain_matrices([T_dest, S_dest, R, S_src, T_src])
#
#    warp_time = utils.timeit(warp_slice)
#    for i in range(10):
#        img_new = warp_time(img, ps, M)[0]
#    f = scroll_plot(img_new[0])
    pos = [150,300,300]
    from elektronn2.utils.plotting import scroll_plot
    from elektronn2.data.image import center_cubes
    warp_time = utils.timeit(warp_slice)

    raw = utils.h5load('~/lustre/mkilling/BirdGT/j0126_old/v2_old_0-raw-zyx.h5')
    if True: # insert stripes
        raw[:,::10,:,:] = 0
        raw[:,:,::20,:] = 0
        raw[:,:,:,::20] = 0

    img = raw

    targ = utils.h5load('~/lustre/mkilling/BirdGT/j0126_old/v2_old_0-combo-sparse-zyx.h5')
    targ[:,::10,:,:] = 0
    targ[:,:,::20,:] = 0
    targ[:,:,:,::20] = 0
    targ = targ

    # --------------------------------
    R = chain_matrices([rotate_y(0.05), rotate_z(0.2)])
    #direc = np.array([1,0.15, 0.05])
    #direc /= np.linalg.norm(direc)
    #direc = [1,0,0]
    #R = get_rotmat_from_direc(direc, 0.2)
    #R[3,1] = 0.002
    ps = [31,131,131]

    z_shift = 0
    aniso_factor = 2
    sample_aniso=True

    dest_center = np.array(ps, dtype=np.float32)/2
    dest_center[0] -= z_shift # sign?
    T_src = translate(-pos[0], -pos[1], -pos[2])
    S_src = scale(aniso_factor, 1, 1)
    if sample_aniso:
        S_dest = scale(1.0/aniso_factor, 1, 1)
    else:
        S_dest = identity()
    T_dest = translate(dest_center[0], dest_center[1], dest_center[2])

    M = chain_matrices([T_dest, S_dest, R, S_src, T_src])

    for i in range(5):
        img_new, targ_new = warp_time(img, ps, M, target=targ, target_ps=[11,121,121], target_discrete_ix=[0,1,2])

    #f0 = scroll_plot(img_new[0])
    i, t = center_cubes(img_new, targ_new)
    f1 = scroll_plot([i[0], t[0]])



# -*- coding: utf-8 -*-
# Compute the MALIS loss function and its derivative wrt the affinity graph
# MAXIMUM spanning tree
# Author: Srini Turaga (sturaga@mit.edu)
# Code integrated to ELEKTRONN2 with permission of author

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, super, zip


import logging
import numpy as np
import scipy.sparse

logger = logging.getLogger('elektronn2log')

try:
    from ._malis import malis_loss_weights, connected_components, marker_watershed
except:
    logger.warning("Malis binary-extension could not be imported/found")

__all__ = ['compute_V_rand_N2',
           'mknhood2d',
           'mknhood3d',
           'mknhood3d_aniso',
           'bmap_to_affgraph',
           'seg_to_affgraph',
           'affgraph_to_seg',
           'malis_weights',
           'watershed_from_affgraph',
           'bmappred_to_affgraph']

def compute_V_rand_N2(seg_true, seg_pred):
    """
    Computes Rand index of ``seg_pred`` w.r.t ``seg_true``.
    Small is better!!!
    The input arrays both contain label IDs and
    may be of arbitrary, but equal, shape.

    Pixels which are have ID in the true segmentation are not counted!

    Parameters:

    seg_true: np.ndarray
      True segmentation, IDs
    seg_pred: np.ndarray
      Predicted segmentation

    Returns
    -------

    ri: ???
    """
    seg_true = seg_true.ravel()
    seg_pred = seg_pred.ravel()
    idx = seg_true != 0
    seg_true = seg_true[idx]
    seg_pred = seg_pred[idx]

    cont_table = scipy.sparse.coo_matrix((np.ones(seg_true.shape),(seg_true,seg_pred))).toarray()
    P = cont_table/cont_table.sum()
    t = P.sum(axis=0)
    s = P.sum(axis=1)

    V_rand_split = (P**2).sum() / (t**2).sum()
    V_rand_merge = (P**2).sum() / (s**2).sum()
    V_rand = 2*(P**2).sum() / ((t**2).sum()+(s**2).sum())

    return (V_rand, V_rand_split, V_rand_merge)

def mknhood2d(radius=1):
    """
    Makes nhood structures for some most used dense graphs
    """
    ceilrad = np.ceil(radius)
    x = np.arange(-ceilrad,ceilrad+1,1)
    y = np.arange(-ceilrad,ceilrad+1,1)
    [i,j] = np.meshgrid(y,x)

    idxkeep = (i**2+j**2)<=radius**2
    i=i[idxkeep].ravel(); j=j[idxkeep].ravel()
    zeroIdx = np.ceil(len(i)/2).astype(np.int32)

    nhood = np.vstack((i[:zeroIdx],j[:zeroIdx])).T.astype(np.int32)
    return np.ascontiguousarray(np.flipud(nhood))

def mknhood3d(radius=1):
    """
    Makes nhood structures for some most used dense graphs.
    The neighborhood reference for the dense graph representation we use
    nhood(1,:) is a 3 vector that describe the node that conn(:,:,:,1) connects to
    so to use it: conn(23,12,42,3) is the edge between node [23 12 42] and [23 12 42]+nhood(3,:)
    See? It's simple! nhood is just the offset vector that the edge corresponds to.
    """

    ceilrad = np.ceil(radius)
    x = np.arange(-ceilrad,ceilrad+1,1)
    y = np.arange(-ceilrad,ceilrad+1,1)
    z = np.arange(-ceilrad,ceilrad+1,1)
    [i,j,k] = np.meshgrid(x,y,z)

    idxkeep = (i**2+j**2+k**2)<=radius**2
    i=i[idxkeep].ravel()
    j=j[idxkeep].ravel()
    k=k[idxkeep].ravel()
    zeroIdx = np.ceil(len(i)/2).astype(np.int32)

    nhood = np.vstack((k[:zeroIdx],i[:zeroIdx],j[:zeroIdx])).T.astype(np.int32)
    return np.ascontiguousarray(np.flipud(nhood))

def mknhood3d_aniso(radiusxy=1,radiusxy_zminus1=1.8):
    """
    Makes nhood structures for some most used dense graphs.
    """

    nhoodxyz = mknhood3d(radiusxy)
    nhoodxy_zminus1 = mknhood2d(radiusxy_zminus1)

    nhood = np.zeros((nhoodxyz.shape[0]+2*nhoodxy_zminus1.shape[0],3),dtype=np.int32)
    nhood[:3,:3] = nhoodxyz
    nhood[3:,0] = -1
    nhood[3:,1:] = np.vstack((nhoodxy_zminus1,-nhoodxy_zminus1))

    return np.ascontiguousarray(nhood)


def nodelist_from_shape(shape, nhood):
    """
    Constructs the two node lists to represent edges of
    an affinity graph for a given volume shape and neighbourhood pattern.

    Parameters
    ----------

    shape: tuple/list
        shape of corresponding volume (z, y, x)
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary

    Returns
    -------

    node1: 4d np.ndarray, int32
        Start nodes, reshaped as array

    node2: 4d np.ndarray, int32
        End nodes, reshaped as array
    """
    n_edge = nhood.shape[0]
    nodes = np.arange(np.prod(shape),dtype=np.int32).reshape(shape)
    node1 = np.tile(nodes, (n_edge,1,1,1))
    node2 = np.full(node1.shape, -1, dtype=np.int32)

    for e in range(n_edge):
        node2[e, \
        max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]), \
        max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]), \
        max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] = \
            nodes[max(0,nhood[e,0]):min(shape[0],shape[0]+nhood[e,0]), \
            max(0,nhood[e,1]):min(shape[1],shape[1]+nhood[e,1]), \
            max(0,nhood[e,2]):min(shape[2],shape[2]+nhood[e,2])]

    return node1, node2



def bmap_to_affgraph(bmap, nhood):
    """
    Construct an affinity graph from a boundary map

    The spatial shape of the affinity graph is the same as of seg_gt.
    This means that some edges are are undefined and therefore treated as disconnected.
    If the offsets in nhood are positive, the edges with largest spatial index are undefined.

    Parameters
    ----------

    bmap: 3d np.ndarray, int
        Volume of boundaries
        0: object interior, 1: boundaries / ECS
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary

    Returns
    -------

    aff: 4d np.ndarray int32
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected

    """
    if np.max(abs(nhood))>1:
        raise ValueError("Cannot construct affinity graph with edges longer\
        than 1 pixel from boundary map. Please construct affinity graph from\
        segmentation / IDs.")

    nhood = np.ascontiguousarray(nhood, np.int32)
    shape = bmap.shape
    n_edge = nhood.shape[0]
    aff = np.ones((n_edge,)+shape,dtype=np.int32)

    for e in range(n_edge):
        aff[e, \
        max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]), \
        max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]), \
        max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] = np.maximum(
            bmap[max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]),
            max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]),
            max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])],
            bmap[max(0,nhood[e,0]):min(shape[0],shape[0]+nhood[e,0]),
            max(0,nhood[e,1]):min(shape[1],shape[1]+nhood[e,1]),
            max(0,nhood[e,2]):min(shape[2],shape[2]+nhood[e,2])] )

    # we have actually accumulated edges which are disconnected by a boundary
    # --> invert
    aff = 1 - aff
    return aff

def bmappred_to_affgraph(pred, nhood):
    """
    Construct an affinity graph from boundary predictions

    Parameters
    ----------

    pred: 3d np.ndarray
        Volume of boundary predictions
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary

    Returns
    -------

    aff: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected
    """
    nhood = np.ascontiguousarray(nhood, np.int32)
    shape = pred.shape
    n_edge = nhood.shape[0]
    if np.issubdtype(pred.dtype, np.int):
        pred = pred.astype(np.float32) / 255

    aff = np.zeros((n_edge,)+shape,dtype=np.float32)

    for e in range(n_edge):
        aff[e,
            max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]),
            max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]),
            max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] = \
            (1 - np.maximum(
              pred[max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]),
                  max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]),
                  max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])],
             pred[max(0,nhood[e,0]):min(shape[0],shape[0]+nhood[e,0]),
                  max(0,nhood[e,1]):min(shape[1],shape[1]+nhood[e,1]),
                  max(0,nhood[e,2]):min(shape[2],shape[2]+nhood[e,2])]) )

    return aff

def seg_to_affgraph(seg_gt, nhood):
    """
    Construct an affinity graph from a segmentation (IDs)

    Segments with ID 0 are regarded as disconnected
    The spatial shape of the affinity graph is the same as of seg_gt.
    This means that some edges are are undefined and therefore treated as disconnected.
    If the offsets in nhood are positive, the edges with largest spatial index are undefined.

    Parameters
    ----------

    seg_gt: 3d np.ndarray, int (any precision)
        Volume of segmentation IDs
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary

    Returns
    -------

    aff: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected
    """
    nhood = np.ascontiguousarray(nhood, np.int32)
    shape = seg_gt.shape
    n_edge = nhood.shape[0]
    aff = np.zeros((n_edge,)+shape,dtype=np.int16)

    for e in range(n_edge):
        aff[e, \
        max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]), \
        max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]), \
        max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] = \
            (seg_gt[max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]),
             max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]),
             max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] ==
             seg_gt[max(0,nhood[e,0]):min(shape[0],shape[0]+nhood[e,0]),
             max(0,nhood[e,1]):min(shape[1],shape[1]+nhood[e,1]),
             max(0,nhood[e,2]):min(shape[2],shape[2]+nhood[e,2])] ) \
            * ( seg_gt[max(0,-nhood[e,0]):min(shape[0],shape[0]-nhood[e,0]),
                max(0,-nhood[e,1]):min(shape[1],shape[1]-nhood[e,1]),
                max(0,-nhood[e,2]):min(shape[2],shape[2]-nhood[e,2])] > 0 ) \
            * ( seg_gt[max(0,nhood[e,0]):min(shape[0],shape[0]+nhood[e,0]),
                max(0,nhood[e,1]):min(shape[1],shape[1]+nhood[e,1]),
                max(0,nhood[e,2]):min(shape[2],shape[2]+nhood[e,2])] > 0 )

    return aff

class AffgraphToSeg(object):
    """
    Parameters
    ----------

    affinity_gt: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    size_thresh: int
        Theshold for size filter after connected components

    Returns
    -------

    seg_gt: 3d np.ndarray, int (any precision)
        Volume of segmentation IDs

    """

    def __init__(self):
        self.edgelist_cache = dict()

    def __call__(self, affinity_gt, nhood, size_thresh=1):
        sh     = affinity_gt.shape
        vol_sh = sh[1:]
        key = (vol_sh, nhood.tobytes()) # cannot hash np.ndarray directly
        if key in self.edgelist_cache:
            node1, node2 = self.edgelist_cache[key]
        else:
            node1, node2 = nodelist_from_shape(vol_sh, nhood)
            node1, node2 = node1.ravel(), node2.ravel()
            self.edgelist_cache[key] = (node1, node2)

        affinity_gt   = np.ascontiguousarray(affinity_gt, dtype=np.float32).ravel()
        size_thresh   = int(size_thresh)

        # CC on GT
        seg_gt, seg_sizes = connected_components(int(np.prod(vol_sh)),
                                                 node1,
                                                 node2,
                                                 affinity_gt,
                                                 size_thresh)
        seg_gt = seg_gt.reshape(vol_sh)

        return seg_gt, seg_sizes

affgraph_to_seg = AffgraphToSeg()



class MalisWeights(object):
    """
    Computes MALIS loss weights

    Roughly speaking the malis weights quantify the impact of an edge in
    the predicted affinity graph on the resulting segmentation.

    Parameters
    ----------
    affinity_pred: 4d np.ndarray float32
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected
    affinity_gt: 4d np.ndarray int16
        Affinity graph of shape (#edges, x, y, z)
        1: connected, 0: disconnected
    seg_gt: 3d np.ndarray, int (any precision)
        Volume of segmentation IDs
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    unrestrict_neg: Bool
        Use this to relax the restriction on neg_counts. The restriction
        modifies the edge weights for before calculating the negative counts
        as: ``edge_weights_neg = np.maximum(affinity_pred, affinity_gt)``
        If unrestricted the predictions are used directly.


    Returns
    -------

    pos_counts: 4d np.ndarray int32
      Impact counts for edges that should be 1 (connect)
    neg_counts: 4d np.ndarray int32
      Impact counts for edges that should be 0 (disconnect)
    """
    def __init__(self):
        self.edgelist_cache = dict()

    def __call__(self, affinity_pred, affinity_gt, seg_gt, nhood,
                 unrestrict_neg=False):

        sh     = affinity_pred.shape
        vol_sh = sh[1:]
        key = (vol_sh, nhood.tobytes()) # cannot hash np.ndarray directly
        if key in self.edgelist_cache:
            node1, node2 = self.edgelist_cache[key]
        else:
            node1, node2 = nodelist_from_shape(vol_sh, nhood)
            node1, node2 = node1.ravel(), node2.ravel()
            self.edgelist_cache[key] = (node1, node2)


        affinity_gt   = affinity_gt.ravel()
        affinity_pred = np.ascontiguousarray(affinity_pred, dtype=np.float32).ravel()
        seg_gt        = np.ascontiguousarray(seg_gt, dtype=np.int32).ravel()

        # MALIS
        edge_weights_pos = np.minimum(affinity_pred, affinity_gt)
        pos_counts = malis_loss_weights(seg_gt,
                                        node1,
                                        node2,
                                        edge_weights_pos,
                                        1)
        if unrestrict_neg:
            edge_weights_neg = affinity_pred
        else:
            edge_weights_neg = np.maximum(affinity_pred, affinity_gt)


        neg_counts = malis_loss_weights(seg_gt,
                                        node1,
                                        node2,
                                        edge_weights_neg,
                                        0)

        pos_counts = pos_counts.reshape(sh)
        neg_counts = neg_counts.reshape(sh)

        return pos_counts, neg_counts

malis_weights = MalisWeights()



def affgraph_to_edgelist(aff, nhood):
    """
    Constructs the two node lists and a list of edge weights to represent edges of
    an affinity graph for a given volume shape and neighbourhood pattern.

    Parameters
    ----------

    shape: tuple/list
     shape of corresponding volume (z, y, x)
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary

    Returns
    -------

    node1: 1d np.ndarray, int
        Start nodes

    node2: 1d np.ndarray, int
        End nodes

    aff: 1d np.ndarray
        Edge weight between node1 and node2
    """
    node1, node2 = nodelist_from_shape(aff.shape[1:], nhood)
    return node1.ravel(), node2.ravel(), aff.ravel()



def watershed_from_affgraph(aff, seeds, nhood):
    node1, node2, edge = affgraph_to_edgelist(aff, nhood)
    marker = np.ascontiguousarray(seeds, dtype=np.int32).ravel()
    seg_gt, seg_sizes = marker_watershed(marker, node1, node2, edge)
    seg_gt = seg_gt.reshape(aff.shape[1:])
    return seg_gt, seg_sizes



if __name__=="__main__":
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    x,y,z = -mknhood3d(1.7).T


    ax.scatter(0,0,0, s=40, c='k')
    ax.scatter(x, y, z)
    ax.set_aspect('equal')
    ax.set_xlim(-2,2)
    ax.set_ylim(-2,2)
    ax.set_zlim(-2,2)
    for i,j,k in zip(x,y,z):
        ax.plot((0,i), (0,j), (0,k))



"""
Visualisiation code taken form Theano
Original Author: Christof Angermueller <cangermueller@gmail.com>
Adapted with permission for the ELEKTRONN2 Toolkit by Marius Killinger 2016
Note that this code is licensed under the original terms of Theano (see license
containing directory).
"""

from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import future.utils
import re
import os
import shutil
import logging
from collections import OrderedDict

from ...neuromancer.graphutils import TaggedShape


logger = logging.getLogger('elektronn2log')


__path__ = os.path.dirname(os.path.realpath(__file__))

pydot_imported = False
try:
    # pydot2 supports py3
    import pydotplus as pd


    if pd.find_graphviz():
        pydot_imported = True
except ImportError:
    try:
        # fall back on pydot if necessary
        import pydot as pd


        if pd.find_graphviz():
            pydot_imported = True
    except ImportError:
        pass  # tests should not fail on optional dependency

if not pydot_imported:
    logger.warning('Failed to import pydot/pydotplus. You must install '
                   'graphviz and either pydot or pydotplus for '
                   '`PyDotFormatter` to work.')


def sort(model, select_outputs):
    graph_sorted = OrderedDict()
    graph_unsorted = OrderedDict()
    for n in model.nodes.values():
        if n.is_source:
            graph_sorted[n] = True
        else:
            graph_unsorted[n] = True

    for n in graph_unsorted:
        if not len(n.children) and select_outputs:  # output node
            ### TODO
            raise NotImplementedError(
                "Select outputs does not work. Make sure "
                "that outputs which are needed from scan "
                "are not removed, that edges to removed "
                "nodes are not drawn and and that after "
                "removing unwanted output nodes in one sort "
                "there will be new unwanted terminal nodes")
            if n.name in select_outputs:
                graph_sorted[n] = True
        else:
            graph_sorted[n] = True

    return graph_sorted


class PyDotFormatter2(object):
    """Create `pydot` graph object from Theano function.

    Parameters
    ----------
    compact : bool
        if True, will remove intermediate variables without name.

    Attributes
    ----------
    node_colors : dict
        Color table of node types.
    apply_colors : dict
        Color table of apply nodes.
    shapes : dict
        Shape table of node types.
    """

    def __init__(self, compact=True):
        """Construct PyDotFormatter object."""
        if not pydot_imported:
            raise ImportError(
                'Failed to import pydot/pydotplus. You must install '
                'graphviz and either pydot or pydotplus for '
                '`PyDotFormatter` to work.')

        self.compact = compact
        self.__node_prefix = 'n'

    def __add_node(self, node):
        """Add new node to node list and return unique id.

        Parameters
        ----------
        node : Theano graph node
            Apply node, tensor variable, or shared variable in compute graph.

        Returns
        -------
        str
            Unique node id.
        """
        assert node not in self.__nodes
        _id = '%s%d' % (self.__node_prefix, len(self.__nodes) + 1)
        self.__nodes[node] = _id
        return _id

    def __node_id(self, node):
        """Return unique node id.

        Parameters
        ----------
        node : Theano graph node
            Apply node, tensor variable, or shared variable in compute graph.

        Returns
        -------
        str
            Unique node id.
        """
        if node in self.__nodes:
            return self.__nodes[node]
        else:
            return self.__add_node(node)

    def get_node_props(self, node):
        cls_name = node.__class__.__name__
        if cls_name.startswith('_'):
            cls_name = cls_name[1:]
        __node_id = self.__node_id(node)

        node_type = 0  # normal
        if node.is_source:
            node_type = 1
        elif len(node.children)==0:
            node_type = 2

        nparams = {}
        nparams['name'] = __node_id
        nparams['label'] = "%s - %s" % (cls_name, node.name)
        nparams['profile'] = [0, 1e-5]
        nparams['style'] = 'filled'
        nparams['type'] = 'colored'

        nparams['shape'] = 'ellipse'
        nparams['fillcolor'] = '#008000'  # 'green'
        if cls_name=='Conv':
            nparams['shape'] = 'invtrapezium'
        if cls_name=='UpConv':
            nparams['shape'] = 'trapezium'
        if cls_name=='Perceptron':
            nparams['shape'] = 'octagon'
        if cls_name in ['GRU', 'LSTM']:
            nparams['shape'] = 'doubleoctagon'
        if cls_name=='Concat':
            nparams['shape'] = 'house'
        if cls_name=='ScanN':
            nparams['shape'] = 'doublecircle'
            nparams['fillcolor'] = 'red'
        if 'loss' in cls_name.lower() or 'nll' in cls_name.lower():
            nparams['shape'] = 'diamond'
            nparams['fillcolor'] = '#FFAA22'
        if cls_name=='ValueNode':
            nparams['shape'] = 'box'

        if node_type==2:
            nparams['fillcolor'] = 'blue'
            nparams['shape'] = 'box'
        elif node_type==1:
            nparams['fillcolor'] = 'yellow'
            nparams['shape'] = 'box'

        if isinstance(node.shape, TaggedShape) and node.shape.ndim:
            nparams['dtype'] = node.shape  # .ext_repr
        elif node.shape:
            nparams['dtype'] = node.shape

        nparams['tag'] = None  # 'tag' # Noone Not needed?
        nparams['node_type'] = 'node type'  # not needed?
        nparams['apply_op'] = 'apply_op'  # not needed?
        return nparams

    def __call__(self, model, select_outputs=None):
        """Create pydot graph from function.

        Parameters
        ----------
        model: model object

        Returns
        -------
        pydot.Dot
            Pydot graph of `fct`
        """
        graph = pd.Dot()
        self.__nodes = {}
        if select_outputs is not None and isinstance(select_outputs, str):
            select_outputs = [select_outputs, ]

        nodes = sort(model, select_outputs)
        # Create nodes
        for node in nodes:
            nparams = self.get_node_props(node)
            pd_node = dict_to_pdnode(nparams)
            graph.add_node(pd_node)

        # Create edges
        for node in nodes:
            for i, c in enumerate(node.children.values()):
                if c.__class__.__name__=="ScanN":
                    if node in c.in_memory:
                        # print("Skippnig",node,'for',c)
                        continue
                p_id = self.__node_id(node)
                c_id = self.__node_id(c)
                edge_params = {}
                edge_params['color'] = 'black'
                edge_label = " "  # str(i)
                pdedge = pd.Edge(p_id, c_id, label=edge_label, **edge_params)
                graph.add_edge(pdedge)

            if node.__class__.__name__=='ScanN':
                self.add_scan_edges(node, graph, nodes)

        return graph

    def add_scan_edges(self, scan, graph, nodes):
        n = str(scan.n_steps) if scan.n_steps else "variable"
        # if scan.out_memory:
        #     out = scan.out_memory
        # else:
        #     out = scan.step_result
        out = []
        for i in scan.out_memory_sl:
            name = scan.output_names[i]
            for nd in nodes:
                if nd.name==name:
                    out.append(nd)
                    break
            else:
                out.append("NotFound")

        # out = [nodes[name] for name in scan.output_name]

        for p, c in zip(out, scan.in_memory):
            for ci in c.children.values():
                if ci.__class__.__name__=="ScanN":
                    continue
                    # if c in c.in_memory:

                p_id = self.__node_id(p)
                c_id = self.__node_id(ci)
                edge_params = {}
                edge_params['color'] = 'red'
                edge_params['constraint'] = False
                edge_params['penwidth'] = 3
                edge_label = n + "x recur.\nreplace %s" % c.name
                pdedge = pd.Edge(p_id, c_id, label=edge_label, **edge_params)
                graph.add_edge(pdedge)

        if scan.in_iterate:
            for p, c in zip(scan.in_iterate, scan.in_iterate_0):
                p_id = self.__node_id(p)
                c_id = self.__node_id(c)
                edge_params = {}
                edge_params['color'] = 'red'
                edge_params['constraint'] = False
                edge_params['penwidth'] = 3
                edge_label = n + "x recur.\niteration"
                pdedge = pd.Edge(p_id, c_id, label=edge_label, **edge_params)
                graph.add_edge(pdedge)


def dict_to_pdnode(d):
    """Create pydot node from dict."""
    e = dict()
    for k, v in d.items():
        if v is not None:
            if isinstance(v, list):
                v = '\t'.join([str(x) for x in v])
            else:
                v = str(v)
            v = str(v)
            v = v.replace('"', '\'')
            e[k] = v
    pynode = pd.Node(**e)
    return pynode


def replace_patterns(x, replace):
    """Replace `replace` in string `x`.

    Parameters
    ----------
    s : str
        String on which function is applied
    replace : dict
        `key`, `value` pairs where key is a regular expression and `value` a
        string by which `key` is replaced
    """
    for from_, to in replace.items():
        x = x.replace(str(from_), str(to))
    return x


def escape_quotes(s):
    """Escape quotes in string.

    Parameters
    ----------
    s : str
        String on which function is applied
    """
    s = re.sub(r'''(['"])''', r'\\\1', s)
    return s


def visualise_model(model, outfile, copy_deps=True, select_outputs=None,
                    image_format='png', *args, **kwargs):
    """
    Parameters
    ----------
    model : model object
    outfile : str
        Path to output HTML file.
    copy_deps : bool, optional
        Copy javascript and CSS dependencies to output directory.

    Notes
    -----
    This function accepts extra parameters which will be forwarded to
    :class:`theano.d3viz.formatting.PyDotFormatter`.

    """

    outfile = os.path.expanduser(outfile)

    # Create DOT graph
    formatter = PyDotFormatter2(*args, **kwargs)
    graph = formatter(model, select_outputs=select_outputs)
    graph.write(outfile + '.' + image_format, prog='dot', format=image_format)

    dot_graph_raw = graph.create_dot()
    if not future.utils.PY2:
        dot_graph_raw = dot_graph_raw.decode('utf8')
    dot_graph = escape_quotes(dot_graph_raw).replace('\n', '').replace('\r',
                                                                       '')

    # Create output directory if not existing
    outdir = os.path.dirname(outfile)
    if not outdir=='' and not os.path.exists(outdir):
        os.makedirs(outdir)

    # Read template HTML file
    template_file = os.path.join(__path__, 'html', 'template.html')
    with open(template_file) as f:
        template = f.read()

    # Copy dependencies to output directory
    src_deps = __path__
    if copy_deps:
        dst_deps = 'd3viz'
        for d in ['js', 'css']:
            dep = os.path.join(outdir, dst_deps, d)
            if not os.path.exists(dep):
                shutil.copytree(os.path.join(src_deps, d), dep)
    else:
        dst_deps = src_deps

    # Replace patterns in template
    replace = {'%% JS_DIR %%': os.path.join(dst_deps, 'js'),
               '%% CSS_DIR %%': os.path.join(dst_deps, 'css'),
               '%% DOT_GRAPH %%': dot_graph,}
    html = replace_patterns(template, replace)

    # Write HTML file
    with open(outfile + '.html', 'w') as f:
        f.write(html)

    graph.write(outfile + '.' + image_format, prog='dot', format=image_format)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import os
from setuptools import setup, find_packages, Extension


malis = False  # Set to True to enable malis build.

ext_modules = []
if malis:
    print('Trying to build the malis extension.\n'
          'Requires a C++ compiler, Numpy, Cython and Boost to be installed.')
    import numpy as np

    ext_modules = [
        Extension(
            "elektronn2.malis._malis",  # Note: _malis_lib.cpp requires boost!
            sources=[
                "elektronn2/malis/_malis.pyx",
                "elektronn2/malis/_malis_lib.cpp",
            ],
            include_dirs=[
                'malis/',
                np.get_include(),
            ],
            language='c++'
        ),
    ]


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="elektronn2",
    version="0.1.0",
    packages=find_packages(),
    scripts=[
        'scripts/elektronn2-train',
    ],
    ext_modules=ext_modules,
    install_requires=[
        'numpy>=1.8',
        'scipy>=0.14',
        'matplotlib>=1.4',
        'h5py>=2.2',
        'theano>=0.7,<0.10',  # ELEKTRONN2 relies on the old Theano backend that has been removed now
        'future>=0.15',
        'tqdm>=4.5',
        'colorlog>=2.7',
        'prompt_toolkit>=1.0.3',
        'jedi>=0.9.0',
        'pydotplus',
        'seaborn',
        'scikit-learn',
        'psutil>=5.0.1',
        'scikit-image>=0.12.3',
        'numba>=0.25'
    ],
    extras_require={
        'knossos': ['knossos_utils'],
        'ipython': ['ipython'],
    },
    include_package_data=True,
    author="Marius Killinger",
    author_email="Marius.Killinger@mailbox.org",
    description="ELEKTRONN2 a is highly configurable toolkit for training 3d/2d CNNs and general Neural Networks",
    long_description=read('README.rst'),
    license="GPL",
    keywords="cnn theano neural network machine learning classification",
    url="http://www.elektronn.org/",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis", ],
)



# -*- coding: utf-8 -*-
# ELEKTRONN2 Toolkit
# Copyright (c) 2015 Marius F. Killinger
# All rights reserved
from __future__ import absolute_import, division, print_function
from builtins import filter, hex, input, int, map, next, oct, pow, range, \
    super, zip

import datetime
import os
import socket
import logging
import numpy as np

from .utils import gpu


logger = logging.getLogger('elektronn2log')

host_name = socket.gethostname()
if host_name in ['synapse03', 'synapse04', 'synapse05', 'synapse06',
                 'synapse07', 'synapse08']:
    cuda_root = "/usr/local/centos-cuda/cuda-6.5"
    logger.info("On Host %s: setting cuda root to %s and disabling DNN!" % (
        host_name, cuda_root))

    import theano

    theano.config.cuda.root = cuda_root
    theano.config.dnn.enabled = "False"

__all__ = ['config', 'change_logging_file']

np.set_printoptions(precision=3, linewidth=90)


def change_logging_file(logger_, save_path, file_name='elektronn2.log'):
    old_lfile_handler = \
        [x for x in logger_.handlers if isinstance(x, logging.FileHandler)][0]

    logger_.removeHandler(old_lfile_handler)

    if not os.path.exists(save_path):
        os.mkdir(save_path)

    lfile_formatter = logging.Formatter(
        '[%(asctime)s] [%(levelname)s]\t%(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')
    lfile_path = os.path.join(save_path, file_name)
    lfile_level = logging.DEBUG
    lfile_handler = logging.FileHandler(lfile_path)
    lfile_handler.setLevel(lfile_level)
    lfile_handler.setFormatter(lfile_formatter)
    logger_.addHandler(lfile_handler)


class DefaultConfig(object):
    def __init__(self):
        """
        This class hard-codes the distribution-wide default values
        """
        ### Toolkit Setup ###------------------------------------------------------
        # (*) <String>: where to create the CNN directory.
        # In this directory a new folder is created with the name of the model
        self.save_path = os.path.abspath(os.path.expanduser("~/CNN_Training/"))
        self.plot_on = True  # <Bool>: whether to create plots of the errors etc.
        self.print_status = True  # <Bool>: whether to print Training status to std.out
        self.device = None  # None/int/'auto' (use .theanorc value) or int (use gpu<i>)
        self.param_save_h = 1.0  # hours: frequency to save a permanent parameter snapshot
        self.param_save_it = 10000
        self.initial_prev_h = 1.0  # hours: time after which first preview is made
        self.prev_save_h = 3.0  # hours: frequency to create previews
        self.overwrite = True  # <Bool>: whether to delete/overwrite existing directory
        # <Bool>/<Int>: whether to "pre-fetch" batches in separate background
        # process, <Bool> or number of processes (True-->2)
        self.background_processes = False
        self.time_per_step_smoothing_length = 50
        self.loss_smoothing_length = 200  #
        self.use_manual_cudnn_conv = True
        self.use_manual_cudnn_conv_not_w1 = True
        self.use_manual_cudnn_pool = True
        self.allow_floatX_downcast = True
        self.show_axis_order_warning = True
        self.use_ortho_init = False
        # Flag used to conditionally do something different for inspection/debugging
        self.inspection = False
        # Whether to create a backup of the current source code in the save directory
        self.backupsrc = True
        self.__doc__ = ""  # Just a hack

        self.read_user_config()

        # Init GPU stuff
        if self.device or self.device==0:
            if self.device=='auto':
                gpu_num = gpu.get_free_gpu(wait=2.0)
                if gpu_num < 0:
                    raise RuntimeError("Could not find free GPU.")

            else:
                assert isinstance(self.device, int)
                gpu_num = self.device

            gpu.initgpu(gpu_num)

        change_logging_file(logger, self.save_path, file_name='elektronn2.log')

    def read_user_config(self):
        config_dict = dict()
        user_path = os.path.abspath(os.path.expanduser('~/.elektronn2rc'))
        if not os.path.exists(user_path):
            logger.debug("No user config file at %s, using default values" % (
                user_path,))
        else:
            try:
                logger.debug("Reading user DefaultConfig from ~/.elektronn2rc")
                exec (compile(open(user_path).read(), user_path, 'exec'), {},
                      config_dict)
            except Exception as e:
                raise RuntimeError("The user config file %s does exist, "
                                   "but an error happened during reading, it might contain "
                                   "invalid code. Error: \n  %s" % (
                                       user_path, e))

            for key in config_dict:
                if key in ['save_path']:
                    config_dict[key] = os.path.abspath(
                        os.path.expanduser(config_dict[key]))
                if key=='param_save_h':
                    raise DeprecationWarning(
                        "Don't use param_save_h, use param_save_it")
                setattr(self, key, config_dict[key])


config = DefaultConfig()
